"""
Base classes for AC/DC devices.
"""

from andes.core import ModelData, Model, IdxParam, NumParam, ExtAlgeb


class ACDC2Term(ModelData, Model):
    """
    AC to two-terminal DC device template.
    """

    def __init__(self, system, config):
        ModelData.__init__(self)
        self.bus = IdxParam(model='Bus',
                            info="idx of connected bus",
                            mandatory=True,
                            )
        self.node1 = IdxParam(default=None,
                              info='Node 1 index',
                              mandatory=True,
                              model='Node',
                              )
        self.node2 = IdxParam(default=None,
                              info='Node 2 index',
                              mandatory=True,
                              model='Node',
                              )
        self.Vn = NumParam(default=110.0,
                           info="AC voltage rating",
                           non_zero=True,
                           tex_name=r'V_n',
                           )
        self.Vdcn1 = NumParam(default=100,
                              info='DC voltage rating on node 1',
                              unit='kV',
                              non_zero=True,
                              tex_name='V_{dcn1}',
                              )
        self.Vdcn2 = NumParam(default=100,
                              info='DC voltage rating on node 2',
                              unit='kV',
                              non_zero=True,
                              tex_name='V_{dcn2}',
                              )
        self.Idcn = NumParam(default=1,
                             info='DC current rating',
                             unit='kA',
                             non_zero=True,
                             tex_name='I_{dcn}',
                             )

        Model.__init__(self, system, config)
        self.a = ExtAlgeb(model='Bus',
                          src='a',
                          indexer=self.bus,
                          info='AC bus voltage phase',
                          )
        self.v = ExtAlgeb(model='Bus',
                          src='v',
                          indexer=self.bus,
                          info='AC bus voltage magnitude',
                          )
        self.v1 = ExtAlgeb(model='Node',
                           src='v',
                           indexer=self.node1,
                           info='DC node 1 voltage',
                           )
        self.v2 = ExtAlgeb(model='Node',
                           src='v',
                           indexer=self.node2,
                           info='DC node 2 voltage',
                           )



from andes.core.model import Model, ModelData
from andes.core.param import ExtParam, IdxParam, NumParam
from andes.core.service import ExtService, VarService
from andes.core.var import Algeb, ExtAlgeb


class IEEEVCData(ModelData):
    """
    IEEEVC data.
    """

    def __init__(self):
        ModelData.__init__(self)
        self.avr = IdxParam(info='Exciter idx', mandatory=True, model='Exciter')

        self.rc = NumParam(default=0.0,
                           info="Active compensation degree.",
                           z=True,
                           tex_name='r_c'
                           )
        self.xc = NumParam(default=0.0,
                           info="Reactive compensation degree.",
                           z=True,
                           tex_name='x_c'
                           )


class IEEEVCModel(Model):
    """
    Implementation of the IEEEVC model.
    """

    def __init__(self, system, config):
        Model.__init__(self, system, config)
        self.group = 'VoltComp'
        self.flags.tds = True

        # retrieve indices of connected generator, bus, and bus freq
        self.syn = ExtParam(model='Exciter', src='syn', indexer=self.avr, export=False,
                            info='Retrieved generator idx', vtype=str)

        self.vf0 = ExtService(src='vf',
                              model='SynGen',
                              indexer=self.syn,
                              tex_name=r'v_{f0}',
                              info='Steady state excitation voltage')
        # from Bus
        self.v = ExtAlgeb(model='SynGen', src='v', indexer=self.syn, tex_name=r'V',
                          info='Retrieved bus terminal voltage',
                          )
        # vd, vq, Id, Iq from SynGen
        self.vd = ExtAlgeb(src='vd',
                           model='SynGen',
                           indexer=self.syn,
                           tex_name=r'V_d',
                           info='d-axis machine voltage',
                           )
        self.vq = ExtAlgeb(src='vq',
                           model='SynGen',
                           indexer=self.syn,
                           tex_name=r'V_q',
                           info='q-axis machine voltage',
                           )
        self.Id = ExtAlgeb(src='Id',
                           model='SynGen',
                           indexer=self.syn,
                           tex_name=r'I_d',
                           info='d-axis machine current',
                           )
        self.Iq = ExtAlgeb(src='Iq',
                           model='SynGen',
                           indexer=self.syn,
                           tex_name=r'I_q',
                           info='q-axis machine current',
                           )

        self.vct = VarService(tex_name=r'V_{CT}',
                              v_str='u * Abs((vd + 1j*vq) + (rc + 1j * xc) * (Id + 1j*Iq))',
                              )

        # output voltage.
        # `vcomp` is the additional voltage to be added to bus terminal voltage
        self.vcomp = Algeb(info='Compensator output voltage to exciter',
                           tex_name=r'v_{comp}',
                           v_str='vct - u * v',
                           e_str='vct - u * v - vcomp',
                           )

        # do not need to interface to exciters here.
        # Let the exciters pick up `vcomp` through back referencing

        self.Eterm = ExtAlgeb(model='Exciter',
                              src='v',
                              indexer=self.avr,
                              v_str='vcomp',
                              e_str='vcomp',
                              )


class IEEEVC(IEEEVCData, IEEEVCModel):
    """
    Voltage compensator IEEEVC model.

    Reference:

    [1] PowerWorld, Voltage Compensator, IEEEVC, [Online],

    [2] NEPLAN, Exciters Models, [Online],

    Available:

    https://www.powerworld.com/WebHelp/Content/TransientModels_HTML/Voltage%20Compensator%20IEEEVC.htm?TocPath=%7C%7C%7CIEEEVC%7C_____0

    https://www.neplan.ch/wp-content/uploads/2015/08/Nep_EXCITERS1.pdf
    """
    def __init__(self, system, config):
        IEEEVCData.__init__(self)
        IEEEVCModel.__init__(self, system, config)



"""
SuiteSparse solvers provided by ``kvxopt``.
"""

import logging

import numpy as np
from kvxopt import matrix, umfpack, klu

logger = logging.getLogger(__name__)


class SuiteSparseSolver:
    """
    Base SuiteSparse solver interface.

    Need to be derived by specific solvers such as UMFPACK or KLU.
    """

    def __init__(self):
        self.A = None
        self.b = None
        self.F = None   # symbolic factorization
        self.N = None   # numeric factorization
        self.factorize = True
        self.use_linsolve = False

    def clear(self):
        """
        Remove all cached PyCapsule of C objects
        """
        self.A = None
        self.b = None
        self.F = None   # symbolic factorization
        self.N = None   # numeric factorization
        self.factorize = True
        self.use_linsolve = False

    def _symbolic(self, A):
        """
        Return the symbolic factorization of sparse matrix ``A``.

        Parameters
        ----------
        A
            Sparse matrix to be factorized.

        Returns
        -------
        A C-object of the symbolic factorization.
        """
        raise NotImplementedError("Method needs to implemented by solver class.")

    def _numeric(self, A, F):
        """
        Return the numeric factorization of sparse matrix ``A`` using symbolic factorization ``F``.

        Parameters
        ----------
        A
            Sparse matrix for the equation set coefficients.
        F
            The symbolic factorization of a matrix with the same non-zero shape as ``A``.

        Returns
        -------
        The numeric factorization of ``A``.
        """
        raise NotImplementedError("Method needs to implemented by solver class.")

    def _solve(self, A, F, N, b):
        """
        Solve linear system ``Ax = b`` using numeric factorization ``N`` and symbolic factorization ``F``.

        Parameters
        ----------
        A
            Sparse matrix.
        F
            Symbolic factorization
        N
            Numeric factorization
        b
            RHS of the equation

        Returns
        -------
        The solution as a ``kvxopt.matrix``.
        """
        raise NotImplementedError("Method needs to implemented by solver class.")

    def solve(self, A, b):
        """
        Solve linear system ``Ax = b`` using numeric factorization ``N`` and symbolic factorization ``F``.
        Store the solution in ``b``.

        This function caches the symbolic factorization in ``self.F`` and is faster in general.
        Will attempt ``Solver.linsolve`` if the cached symbolic factorization is invalid.

        Parameters
        ----------
        A
            Sparse matrix for the equation set coefficients.
        F
            The symbolic factorization of A or a matrix with the same non-zero shape as ``A``.
        N
            Numeric factorization of A.
        b
            RHS of the equation.

        Returns
        -------
        numpy.ndarray
            The solution in a 1-D ndarray
        """
        self.A = A
        self.b = b

        if self.factorize is True:
            self.F = self._symbolic(self.A)
            self.factorize = False

        try:
            self.N = self._numeric(self.A, self.F)
            self._solve(self.A, self.F, self.N, self.b)

            return np.ravel(self.b)
        except ValueError:
            logger.debug('Unexpected symbolic factorization.')
            self.F = self._symbolic(self.A)
            self.solve(self.A, self.b)

            return np.ravel(self.b)
        except ArithmeticError:
            logger.error('Jacobian matrix is singular.')
            # diag = self.A[0:self.A.size[0] ** 2:self.A.size[0]+1]
            # idx = (np.argwhere(np.array(matrix(diag)).ravel() == 0.0)).ravel()
            # logger.error('The xy indices of associated variables:')
            # logger.error(' '.join([str(item) for item in idx]))

            # works around a KVXOPT bug
            suspect_diag = []
            for i in range(self.A.size[0]):
                if self.A[i, i] == 0.0:
                    suspect_diag.append(i)

            logger.error('Suspect diagonal elements: {}'.format(suspect_diag))

            return np.ravel(matrix(np.nan, self.b.size, 'd'))

    def linsolve(self, A, b):
        """
        Solve linear equation set ``Ax = b`` and returns the solutions in a 1-D array.

        This function performs both symbolic and numeric factorizations every time, and can be slower than
        ``Solver.solve``.

        Parameters
        ----------
        A
            Sparse matrix

        b
            RHS of the equation

        Returns
        -------
        The solution in a 1-D np array.
        """
        raise NotImplementedError


class UMFPACKSolver(SuiteSparseSolver):
    """
    UMFPACK solver.

    Utilizes ``kvxopt.umfpack`` for factorization.
    """

    def __init__(self):
        super().__init__()

    def _symbolic(self, A):
        return umfpack.symbolic(A)

    def _numeric(self, A, F):
        return umfpack.numeric(A, F)

    def _solve(self, A, F, N, b):
        umfpack.solve(A, N, b)

    def linsolve(self, A, b):
        try:
            umfpack.linsolve(A, b)
        except ArithmeticError:
            logger.error('Singular matrix. Case is not solvable')
        return np.ravel(b)


class KLUSolver(SuiteSparseSolver):
    """
    KLU solver.
    """

    def __init__(self):
        super().__init__()

    def _symbolic(self, A):
        return klu.symbolic(A)

    def _numeric(self, A, F):
        return klu.numeric(A, F)

    def _solve(self, A, F, N, b):
        klu.solve(A, F, N, b)

    def linsolve(self, A, b):
        try:
            klu.linsolve(A, b)
        except ArithmeticError:
            logger.error('Singular matrix. Case is not solvable')
        return np.ravel(b)



"""
Phase measurement loop model.
"""
from andes.core.model import ModelData, Model
from andes.core.param import NumParam, IdxParam
from andes.core.block import PIController, Lag  # NOQA
from andes.core.var import ExtAlgeb, State


class PLL1Data(ModelData):
    """
    Data for PLL.
    """

    def __init__(self):
        ModelData.__init__(self)

        self.bus = IdxParam(info="bus idx", mandatory=True)

        self.Kp = NumParam(info='proportional gain', default=1,
                           tex_name='K_p',
                           )

        self.Ki = NumParam(info='integral gain', default=0.2,
                           tex_name='K_i',
                           )

        self.Tf = NumParam(default=0.05,
                           info="input digital filter time const",
                           unit="sec",
                           tex_name='T_f',
                           )
        self.Tp = NumParam(default=0.05,
                           info='output filter time const.',
                           unit='sec',
                           tex_name='T_p')

        self.fn = NumParam(default=60.0,
                           info="nominal frequency",
                           unit='Hz',
                           tex_name='f_n',
                           )


class PLL1Model(Model):
    """
    Simple PLL1 implementation.
    """

    def __init__(self, system, config):
        super().__init__(system, config)

        self.flags.tds = True
        self.group = 'PLL'

        self.a = ExtAlgeb(model='Bus',
                          src='a',
                          indexer=self.bus,
                          tex_name=r'\theta',
                          info='Bus voltage angle'
                          )

        self.af = Lag(u=self.a, T=self.Tf, K=1, D=1,
                      info='input angle signal filter',
                      )

        self.PI = PIController(u='u * (af_y - am)', kp=self.Kp, ki=self.Ki,
                               tex_name='PI',
                               info='PI controller',
                               )

        self.ae = State(info='PLL angle output before filter',
                        e_str='2 * pi *fn * PI_y', v_str='a',
                        tex_name=r'\theta_{est}'
                        )

        self.am = State(info='PLL output angle after filtering',
                        e_str='ae - am',
                        t_const=self.Tp,
                        v_str='a',
                        tex_name=r'\theta_{PLL}'
                        )


class PLL1(PLL1Data, PLL1Model):
    """
    Simple Phasor Lock Loop (PLL) using one PI controller.

    Input bus angle signal -> Lag filter 1 with Tf ->
    PI Controller (Kp, Ki) -> Estimated angle (2 * pi * fn * PI_y) ->
    Lag filter 2 with Tp

    The output signal is ``am``.
    """

    def __init__(self, system, config):
        PLL1Data.__init__(self)
        PLL1Model.__init__(self, system, config)



"""Distributed energy storage system model"""

from andes.core.block import Integrator
from andes.core.discrete import HardLimiter, LessThan
from andes.core.param import NumParam
from andes.core.var import Algeb, AliasState
from andes.models.distributed.pvd1 import PVD1Data, PVD1Model


class ESD1Data(PVD1Data):
    """
    Data for energy storage distributed model.
    """

    def __init__(self):
        PVD1Data.__init__(self)
        self.Tf = NumParam(default=1.0, tex_name='T_f',
                           info='Integrator constant for SOC model',
                           )
        self.SOCmin = NumParam(default=0.0, tex_name='SOC_{min}',
                               info='Minimum required value for SOC in limiter',
                               )

        self.SOCmax = NumParam(default=1.0, tex_name='SOC_{max}',
                               info='Maximum allowed value for SOC in limiter',
                               )

        self.SOCinit = NumParam(default=0.5, tex_name='SOC_{init}',
                                info='Initial state of charge',
                                )

        self.En = NumParam(default=100.0, tex_name='E_n',
                           info='Rated energy capacity',
                           unit="MWh"
                           )

        self.EtaC = NumParam(default=1.0, tex_name='Eta_C',
                             info='Efficiency during charging',
                             vrange=(0, 1),
                             )

        self.EtaD = NumParam(default=1.0, tex_name='Eta_D',
                             info='Efficiency during discharging',
                             vrange=(0, 1),
                             )


class ESD1Model(PVD1Model):
    """
    Model implementation of ESD1.
    """

    def __init__(self, system, config):
        PVD1Model.__init__(self, system, config)

        # --- Determine whether the energy storage is in charging or discharging mode ---
        self.LTN = LessThan(self.Ipout_y, 0.0)

        # --- Add integrator. Assume that state-of-charge is the initial condition ---
        self.pIG = Integrator(u='-LTN_z1*(v * Ipout_y)*EtaC - LTN_z0*(v * Ipout_y)/EtaD',
                              T=self.Tf, K='sys_mva / 3600 / En', y0=self.SOCinit,
                              check_init=False,
                              )
        self.pIG.info = 'State of charge'

        self.SOC = AliasState(self.pIG_y)
        self.SOC.info = 'Alias for state of charge'

        # --- Add hard limiter for SOC ---
        self.SOClim = HardLimiter(u=self.pIG_y, lower=self.SOCmin, upper=self.SOCmax)

        # --- Add Ipmax, Ipmin, and Ipcmd ---
        self.Ipmax.v_str = '(1-SOClim_zl)*(SWPQ_s1 * ialim + SWPQ_s0 * sqrt(Ipmaxsq0))'
        self.Ipmax.e_str = '(1-SOClim_zl)*(SWPQ_s1 * ialim + SWPQ_s0 * sqrt(Ipmaxsq)) - Ipmax'

        self.Ipmin = Algeb(info='Minimum value of Ip',
                           v_str='-(1-SOClim_zu) * (SWPQ_s1 * ialim + SWPQ_s0 * sqrt(Ipmaxsq0))',
                           e_str='-(1-SOClim_zu) * (SWPQ_s1 * ialim + SWPQ_s0 * sqrt(Ipmaxsq)) - Ipmin',
                           )

        self.Ipcmd.lim.lower = self.Ipmin
        self.Ipcmd.y.deps = ['Ipmin']


class ESD1(ESD1Data, ESD1Model):
    """
    Distributed energy storage model.

    A state-of-charge limit is added to the PVD1 model.
    This limit is applied to Ipmax and Ipmin.
    The state of charge is in state variable ``SOC``,
    which is an alias of ``pIG_y``.

    Reference:
    [1] Powerworld, Renewable Energy Electrical Control Model REEC_C
    Available:

    https://www.powerworld.com/WebHelp/Content/TransientModels_HTML/Exciter%20REEC_C.htm
    """
    def __init__(self, system, config):
        ESD1Data.__init__(self)
        ESD1Model.__init__(self, system, config)



"""
Simple PMU based on low-pass filter.
"""

from andes.core import ModelData, IdxParam, NumParam, Model, ExtAlgeb, State


class PMUData(ModelData):
    """
    Phasor measurement unit data.
    """

    def __init__(self):
        ModelData.__init__(self)
        self.bus = IdxParam(info="bus idx", mandatory=True)

        self.Ta = NumParam(default=0.1, tex_name='T_a', info='angle filter time constant')
        self.Tv = NumParam(default=0.1, tex_name='T_v', info='voltage filter time constant')


class PMU(PMUData, Model):
    """
    Simple phasor measurement unit model.

    This model tracks the bus voltage magnitude and phase angle, each using
    a low-pass filter.
    """

    def __init__(self, system, config):
        PMUData.__init__(self)
        Model.__init__(self, system, config)

        self.flags.tds = True
        self.group = 'PhasorMeasurement'

        self.a = ExtAlgeb(model='Bus',
                          src='a',
                          indexer=self.bus,
                          tex_name=r'\theta',
                          info='Bus voltage phase angle',
                          )
        self.v = ExtAlgeb(model='Bus',
                          src='v',
                          indexer=self.bus,
                          tex_name=r'V',
                          info='Bus voltage magnitude',
                          )

        self.am = State(tex_name=r'\theta_m', info='phase angle measurement',
                        unit='rad.', e_str='a - am', t_const=self.Ta, v_str='a',
                        )

        self.vm = State(tex_name='V_m', info='voltage magnitude measurement',
                        unit='p.u.(kV)', e_str='v - vm', t_const=self.Tv, v_str='v',
                        )



"""
JSON reader and writer for ANDES.

"""
import io
import json
import logging

from collections import OrderedDict
from typing import Union

from andes.utils.paths import confirm_overwrite

logger = logging.getLogger(__name__)


def testlines(infile):
    return True


def write(system, outfile, skip_empty=True, overwrite=None, **kwargs):
    """
    Write loaded ANDES system data into a JSON file.

    Parameters
    ----------
    system : System
        A loaded system with parameters
    outfile : str
        Path to the output file
    skip_empty : bool
        Skip output of empty models (n = 0)
    overwrite : bool
        None to prompt for overwrite selection; True to overwrite; False to not overwrite

    Returns
    -------
    bool
        True if file written; False otherwise
    """
    if not confirm_overwrite(outfile, overwrite):
        return False

    if hasattr(outfile, 'write'):
        outfile.write(_dump_system(system, skip_empty))
    else:
        with open(outfile, 'w') as writer:
            writer.write(_dump_system(system, skip_empty))
            logger.info('JSON file written to "%s"', outfile)

    return True


def _dump_system(system, skip_empty, orient='records'):
    """
    Dump parameters of each model into a json string and return
    them all in an OrderedDict.
    """

    out = OrderedDict()
    for name, instance in system.models.items():
        if skip_empty and instance.n == 0:
            continue
        out[name] = instance.cache.df_in.to_dict(orient=orient)

    return json.dumps(out, indent=2)


def read(system, infile: Union[str, io.IOBase]):
    """
    Read JSON file with ANDES model data into an empty system.

    Parameters
    ----------
    system : System
        Empty System instance
    infile : str or io.BaseIO
        str: path to the input file; or io.BaseIO: a stream to
        read from

    Returns
    -------
    System
        System instance after succeeded
    """
    if isinstance(infile, str):
        f = open(infile, 'r')
    else:
        f = infile

    json_in = json.load(f)

    if f is not infile:
        f.close()

    for name, dct in json_in.items():
        for row in dct:
            system.add(name, row)

    return system



#  [ANDES] (C)2015-2022 Hantao Cui
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  File name: var.py
#  Last modified: 8/16/20, 7:27 PM

from typing import List, Optional, Union

from andes.core.common import DummyValue
from andes.core.discrete import Discrete
from andes.core.param import BaseParam
from andes.core.service import BaseService
from andes.models.group import GroupBase
from andes.shared import np


class BaseVar:
    """
    Base variable class.

    Derived classes `State` and `Algeb` should be used to build
    model variables.

    Parameters
    ----------
    name : str, optional
        Variable name
    info : str, optional
        Descriptive information
    unit : str, optional
        Unit
    tex_name : str
        LaTeX-formatted variable name. If is None, use `name`
        instead.
    discrete : Discrete
        Discrete component on which thi variable depends on.
        ANDES will call `check_var()` of the discrete component
        before initializing this variable.

    Attributes
    ----------
    a : array-like
        variable address
    v : array-like
        local-storage of the variable value
    e : array-like
        local-storage of the corresponding equation value
    e_str : str
        the string/symbolic representation of the equation
    v_str : str
        explicit initialization equation
    v_str_add : bool
        True if the value of `v_str` will be added to the variable.
        Useful when other models access this variable and set part
        of the initial value
    v_iter : str
        implicit iterative equation in the form of 0 = v_iter
    """

    def __init__(self,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 v_str: Optional[Union[str, float]] = None,
                 v_iter: Optional[str] = None,
                 e_str: Optional[str] = None,
                 discrete: Optional[Discrete] = None,
                 v_setter: Optional[bool] = False,
                 e_setter: Optional[bool] = False,
                 v_str_add: Optional[bool] = False,
                 addressable: Optional[bool] = True,
                 export: Optional[bool] = True,
                 diag_eps: Optional[float] = 0.0,
                 deps: Optional[List] = None,
                 is_output: Optional[bool] = False,
                 ):

        self.name = name
        self.info = info
        self.unit = unit

        self.tex_name = tex_name if tex_name else name
        self.owner = None  # instance of the owner Model
        self.id = None     # variable internal index inside a model (assigned in run time)

        self.v_str = v_str    # equation string (v = v_str) for variable initialization
        self.v_iter = v_iter  # the implicit equation (0 = v_iter) for iterative initialization
        self.e_str = e_str    # equation string

        self.discrete = discrete
        self.v_setter = v_setter        # True if this variable sets the variable value
        self.e_setter = e_setter        # True if this var sets the equation value
        self.v_str_add = v_str_add

        self.addressable = addressable  # True if this var needs to be assigned an address FIXME: not in use
        self.export = export            # True if this var's value needs to exported
        self.diag_eps = diag_eps        # small diagonal value to be added to `dae.gy`
        self.deps = deps          # a list of variable names this BaseVar depends on for initialization
        self.is_output = is_output      # indicate if this variable is an output terminal
        self.is_input = False     # internal variables are never inputs

        # --- attributes assigned by `set_address` begins ---
        self.n = 0

        # address into the variable and equation arrays (dae.f/dae.g and dae.x/dae.y)
        self.a: np.ndarray = np.array([], dtype=int)

        # address into external equation RHS array (dae.h/dae.i)
        self.r: np.ndarray = np.array([], dtype=int)

        self.av: np.ndarray = np.array([], dtype=int)      # FIXME: future var. address array
        self.ae: np.ndarray = np.array([], dtype=int)      # FIXME: future equation address array
        # --- attributes assigned by `set_address` ends ---

        self.v: np.ndarray = np.array([], dtype=float)  # variable value array
        self.e: np.ndarray = np.array([], dtype=float)  # equation value array

        # internal flags
        # NOTE:
        # contiguous is True only for internal variables of models with flag `collate = False`.
        self._contiguous = False  # True if if address is contiguous to allow slicing into arrays without copy.

        self.e_inplace = False    # True if `self.e` is in-place access to `System.dae.__dict__[self.e_code]`
        self.v_inplace = False    # True if `self.v` is in-place access to `System.dae.__dict__[self.v_code]`
        self.allow_none = False   # True to allow None in address (NOT IN USE)

    def reset(self):
        """
        Reset the internal numpy arrays and flags.
        """
        self.n = 0
        self.a[:] = 0
        self.v[:] = 0
        self.e[:] = 0
        self.av[:] = 0
        self.ae[:] = 0

        self._contiguous = False
        self.e_inplace = False
        self.v_inplace = False

    def __repr__(self):
        if self.n == 0:
            span = []

        elif 1 <= self.n <= 20:
            span = f'a={self.a}, v={self.v}, e={self.e}'

        else:
            span = []
            if not isinstance(self, ExtVar):
                span.append(self.a[0])
                span.append(self.a[-1])
                span.append(self.a[1] - self.a[0])
                span = ':'.join([str(i) for i in span])
                span = 'a=[' + span + ']'

        return f'{self.__class__.__name__}: {self.owner.__class__.__name__}.{self.name}, {span}'

    def set_address(self, addr: np.ndarray, contiguous=False):
        """
        Set the address of internal variables.

        Parameters
        ----------
        addr : np.ndarray
            The assigned address for this variable
        contiguous : bool, optional
            If the addresses are contiguous
        """
        self.a = addr
        self.n = len(self.a)

        # NOT IN USE
        self.ae = np.array(self.a)
        self.av = np.array(self.a)
        # -----------

        self._contiguous = contiguous

        if self._contiguous:
            if self.e_setter is False:
                self.e_inplace = True

            if self.v_setter is False:
                self.v_inplace = True

    def set_arrays(self, dae, inplace=True, alloc=True):
        """
        Set the equation and values arrays.


        Parameters
        ----------
        dae : DAE
            Reference to System.dae
        """
        if inplace is True:
            self._set_arrays_inplace(dae)
        if alloc is True:
            self._set_arrays_alloc()

    def _set_arrays_inplace(self, dae):
        """
        Set arrays that share memory with the dae arrays.

        It slicing into DAE due to the contiguous indices.
        """

        slice_idx = slice(self.a[0], self.a[-1] + 1)
        if self.v_inplace:
            self.v = dae.__dict__[self.v_code][slice_idx]

        if self.e_inplace:
            self.e = dae.__dict__[self.e_code][slice_idx]

    def _set_arrays_alloc(self):
        """
        Allocate for internal v and e arrays that cannot
        share memory with dae arrays.
        """

        if not self.v_inplace:
            self.v = np.zeros(self.n)

        if not self.e_inplace:
            self.e = np.zeros(self.n)

    def get_names(self):
        return [self.name]

    @property
    def class_name(self):
        return self.__class__.__name__


class Algeb(BaseVar):
    """
    Algebraic variable class, an alias of the `BaseVar`.

    Attributes
    ----------
    e_code : str
        Equation code string, equals string literal ``g``
    v_code : str
        Variable code string, equals string literal ``y``
    """
    e_code = 'g'
    v_code = 'y'


class State(BaseVar):
    """
    Differential variable class, an alias of the `BaseVar`.

    Parameters
    ----------
    t_const : BaseParam, DummyValue
        Left-hand time constant for the differential equation.
        Time constants will not be evaluated as part of the differential equation.
        They will be collected to array `dae.Tf` to multiply to the right-hand side `dae.f`.
    check_init : bool
        True to check if the equation right-hand-side is zero
        initially. Disabling the checking can be used for integrators
        when the initial input may not be zero.

    Attributes
    ----------
    e_code : str
        Equation code string, equals string literal ``f``
    v_code : str
        Variable code string, equals string literal ``x``
    """
    e_code = 'f'
    v_code = 'x'

    def __init__(self,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 v_str: Optional[Union[str, float]] = None,
                 v_iter: Optional[str] = None,
                 e_str: Optional[str] = None,
                 discrete: Optional[Discrete] = None,
                 t_const: Optional[Union[BaseParam, DummyValue, BaseService]] = None,
                 check_init: Optional[bool] = True,
                 v_setter: Optional[bool] = False,
                 e_setter: Optional[bool] = False,
                 addressable: Optional[bool] = True,
                 export: Optional[bool] = True,
                 diag_eps: Optional[float] = 0.0,
                 deps: Optional[List] = None,
                 ):
        BaseVar.__init__(self, name=name,
                         tex_name=tex_name,
                         info=info,
                         unit=unit,
                         v_str=v_str,
                         v_iter=v_iter,
                         e_str=e_str,
                         discrete=discrete,
                         v_setter=v_setter,
                         e_setter=e_setter,
                         addressable=addressable,
                         export=export,
                         diag_eps=diag_eps,
                         deps=deps,
                         )
        self.t_const = t_const
        self.check_init = check_init


class ExtVar(BaseVar):
    """
    Externally defined algebraic variable

    This class is used to retrieve the addresses of externally-
    defined variable. The `e` value of the `ExtVar` will be added
    to the corresponding address in the DAE equation.

    Parameters
    ----------
    model : str
        Name of the source model
    src : str
        Source variable name
    indexer : BaseParam, BaseService
        A parameter of the hosting model, used as indices into
        the source model and variable. If is None, the source
        variable address will be fully copied.
    allow_none : bool
        True to allow None in indexer

    Attributes
    ----------
    parent_model : Model
        The parent model providing the original parameter.
    uid : array-like
        An array containing the absolute indices into the
        parent_instance values.
    e_code : str
        Equation code string; copied from the parent instance.
    v_code : str
        Variable code string; copied from the parent instance.
    """

    def __init__(self,
                 model: str,
                 src: str,
                 indexer: Optional[Union[List, np.ndarray, BaseParam, BaseService]] = None,
                 allow_none: Optional[bool] = False,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 ename: Optional[str] = None,
                 tex_ename: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 v_str: Optional[Union[str, float]] = None,
                 v_iter: Optional[str] = None,
                 e_str: Optional[str] = None,
                 v_setter: Optional[bool] = False,
                 e_setter: Optional[bool] = False,
                 addressable: Optional[bool] = True,
                 export: Optional[bool] = True,
                 diag_eps: Optional[float] = 0.0,
                 is_input: Optional[bool] = False,
                 ):
        super().__init__(name=name,
                         tex_name=tex_name,
                         info=info,
                         unit=unit,
                         v_str=v_str,
                         v_iter=v_iter,
                         e_str=e_str,
                         v_setter=v_setter,
                         e_setter=e_setter,
                         addressable=addressable,
                         export=export,
                         diag_eps=diag_eps,
                         )
        self.ename = ename  # equation name corresponding to this variable
        self.tex_ename = tex_ename if tex_ename else ename

        self.model = model
        self.src = src
        self.indexer = indexer
        self.allow_none = allow_none
        self.is_input = is_input  # if this ExtVar is an input terminal
        self.is_output = False    # external variables are never outputs

        self.parent = None
        self._idx = None
        self._n = []
        self._n_count = 0

    @property
    def _v(self):
        out = []
        idx = 0
        for n in self._n:
            out.append(self.v[idx:idx+n])
            idx += n
        return out

    @property
    def _a(self):
        out = []
        idx = 0
        for n in self._n:
            out.append(self.a[idx:idx+n])
            idx += n
        return out

    def set_address(self, addr, contiguous=False):
        """
        Assigns address for equation RHS.
        """
        self.r = addr

    def set_arrays(self, dae, inplace=True, alloc=True):
        """
        Access ``dae.h`` or ``dae.i`` for the RHS of external variables
        when ``e_str`` exists..
        """

        if self.e_str is None or (self.n == 0):
            return

        try:
            slice_idx = slice(self.r[0], self.r[-1] + 1)
        except IndexError as e:
            print(self.owner.class_name)
            print(self.name)
            raise e

        if isinstance(self, ExtState):
            self.e = dae.h[slice_idx]
        elif isinstance(self, ExtAlgeb):
            self.e = dae.i[slice_idx]
        else:
            raise NotImplementedError

    def link_external(self, ext_model):
        """
        Update variable addresses provided by external models

        This method sets attributes including `parent_model`,
        `parent_instance`, `uid`, `a`, `n`, `e_code` and
        `v_code`. It initializes the `e` and `v` to zero.

        Returns
        -------
        None

        Parameters
        ----------
        ext_model : Model
            Instance of the parent model

        Warnings
        --------
        `link_external` does not check if the ExtVar type is the same
        as the original variable to reduce performance overhead.
        It will be a silent error (a dimension too small error from `dae.build_pattern`)
        if a model uses `ExtAlgeb` to access a `State`, or vice versa.

        """

        self.parent = ext_model

        if isinstance(ext_model, GroupBase):
            # determine the number of elements based on `indexer.v`
            if self.indexer.n > 0 and isinstance(self.indexer.v[0], (list, np.ndarray)):
                self._n = [len(i) for i in self.indexer.v]  # number of elements in each sublist
                self._idx = np.concatenate([np.array(i) for i in self.indexer.v])
            else:
                self._n = [len(self.indexer.v)]
                self._idx = self.indexer.v

            # use `0` for non-existent addresses (corr. to None in indexer)
            self.a = ext_model.get(src=self.src,
                                   idx=self._idx,
                                   attr='a',
                                   allow_none=self.allow_none,
                                   default=0,
                                   ).astype(int)

            # check if source var type is the same as this ExtVar
            vcodes = np.array(ext_model.get_field(src=self.src, idx=self._idx, field='v_code'))
            vcodes = vcodes[vcodes != np.array(None)].astype(str)

            if not all(vcodes == np.array(self.v_code)):
                raise TypeError("ExtVar <%s.%s> is of type <%s>, but source Vars <%s.%s> may not." %
                                (self.owner.class_name, self.name, self.v_code,
                                 ext_model.class_name, self.src))

            self.n = len(self.a)

        else:
            original_var = ext_model.__dict__[self.src]

            if self.allow_none:
                raise NotImplementedError(f"{self.name}: allow_none not implemented for Model")
            if original_var.v_code != self.v_code:
                raise TypeError("Linking %s of %s to %s of %s is not allowed" %
                                (self.name, self.class_name,
                                 original_var.name, original_var.class_name))

            if self.indexer is not None:
                uid = ext_model.idx2uid(self.indexer.v)
            else:
                uid = np.arange(ext_model.n, dtype=int)

            self._n = [len(uid)]
            if len(uid) > 0:
                self.a = original_var.a[uid]
            else:
                self.a = np.array([], dtype=int)

            # set initial v and e values to zero
            self.n = len(self.a)

        self.v = np.zeros(self.n)
        # `self.e` is assigned in `set_arrays()`


class ExtState(ExtVar):
    """
    External state variable type.

    Warnings
    --------
    ``ExtState`` is not allowed to set ``t_const``, as it will conflict with the
    source ``State`` variable. In fact, one should not set ``e_str`` for ``ExtState``.
    """
    e_code = 'f'
    r_code = 'h'
    v_code = 'x'
    t_const = None


class ExtAlgeb(ExtVar):
    """
    External algebraic variable type.
    """
    e_code = 'g'
    r_code = 'i'
    v_code = 'y'


class AliasAlgeb(ExtAlgeb):
    """
    Alias algebraic variable. Essentially ``ExtAlgeb`` that links to a a model's
    own variable.

    ``AliasAlgeb`` is useful when the final output of a model is from a block,
    but the model must provide the final output in a pre-defined name. Using
    ``AliasAlgeb``, A model can avoid adding an additional variable with a dummy
    equations.

    Like ``ExtVar``, labels of ``AliasAlgeb`` will not be saved in the final
    output. When plotting from file, one need to look up the original variable
    name.
    """

    def __init__(self, var, **kwargs):
        ExtAlgeb.__init__(self,
                          model=var.owner.class_name,
                          src=var.name,
                          indexer=var.owner.idx,
                          info=f'Alias of {var.name}',
                          is_input=False,
                          **kwargs,
                          )


class AliasState(ExtState):
    """
    Alias state variable.

    Refer to the docs of ``AliasAlgeb``.
    """

    def __init__(self, var, **kwargs):
        ExtState.__init__(self,
                          model=var.owner.class_name,
                          src=var.name,
                          indexer=var.owner.idx,
                          info=f'Alias of {var.name}',
                          is_input=False,
                          **kwargs,
                          )



"""
Slack generator model for steady state.
"""

from collections import OrderedDict

from andes.core import NumParam, ExtParam, SortedLimiter, Algeb
from andes.models.static.pv import PVData, PVModel


class SlackData(PVData):
    def __init__(self):
        super().__init__()
        self.a0 = NumParam(default=0.0,
                           info="reference angle set point",
                           tex_name=r'\theta_0')


class Slack(SlackData, PVModel):
    """
    Slack generator.
    """

    def __init__(self, system=None, config=None):
        SlackData.__init__(self)
        PVModel.__init__(self, system, config)

        self.config.add(OrderedDict((('av2pv', 0),
                                     )))
        self.config.add_extra("_help",
                              av2pv="convert Slack to PV in PFlow at P limits",
                              )
        self.config.add_extra("_alt",
                              av2pv=(0, 1),
                              )
        self.config.add_extra("_tex",
                              av2pv="z_{av2pv}",
                              )
        self.busa0 = ExtParam(model='Bus', src='a0', indexer=self.bus,
                              export=False, tex_name=r'\theta_{0bus}',
                              )
        self.a.v_setter = True
        self.a.v_str = 'u * a0 + (1-u) * busa0'

        delattr(self, 'p')
        self.p = Algeb(info='actual active power generation',
                       unit='p.u.',
                       tex_name=r'p',
                       diag_eps=True,
                       v_str='u * p0',
                       )

        self.p.e_str = "u*(plim_zi * (a0-a) + " \
                       "plim_zl * (pmin-p) + " \
                       "plim_zu * (pmax-p))"
        self.a.e_str = '-u * p'

        self.plim = SortedLimiter(u=self.p, lower=self.pmin, upper=self.pmax,
                                  enable=self.config.av2pv)



"""
Distributed energy resource protection model base.
"""
from andes.core.block import IntegratorAntiWindup
from andes.core.discrete import Limiter
from andes.core.model import Model, ModelData
from andes.core.param import ExtParam, IdxParam, NumParam
from andes.core.service import ConstService, ExtendedEvent, ExtService
from andes.core.var import Algeb, ExtAlgeb


class DGPRCTBaseData(ModelData):
    """
    DGPRCT Base model data.
    """

    def __init__(self):
        super(DGPRCTBaseData, self).__init__()

        self.dev = IdxParam(info='idx of the target device',
                            mandatory=True,
                            )

        self.busfreq = IdxParam(model='BusFreq',
                                info='Target device interface bus measurement device idx',
                                )

        # -- protection enable parameters
        self.fen = NumParam(default=1,
                            tex_name='fen',
                            vrange=(0, 1),
                            info='Frequency deviation protection enable. \
                                  1 for enable, 0 for disable.',
                            )

        # TODO: add `fflag` for choice that whether block source signal
        # self.fflag = NumParam(default=1,
        #                       tex_name='fflag',
        #                       vrange=[0, 1],
        #                       info='Frequency lock option. \
        #                             1 to enable freq after protection, \
        #                             0 to disable freq after protection.',
        #                       )

        self.Ven = NumParam(default=0,
                            tex_name='Ven',
                            vrange=(0, 1),
                            info='Voltage deviation protection enable.\
                                  1 for enable, 0 for disable.',
                            )

        # -- protection parameters, frequency
        self.fl3 = NumParam(default=50,
                            tex_name='fl3',
                            info='Under frequency shadding point 3',
                            unit='Hz',
                            )

        self.fl2 = NumParam(default=57.5,
                            tex_name='fl2',
                            info='Over frequency shadding point 2',
                            unit='Hz',
                            )

        self.fl1 = NumParam(default=59.2,
                            tex_name='fl1',
                            info='Under frequency shadding point 1',
                            unit='Hz',
                            )

        self.fu1 = NumParam(default=60.5,
                            tex_name='fu1',
                            info='Over frequency shadding point 1',
                            unit='Hz',
                            )

        self.fu2 = NumParam(default=61.5,
                            tex_name='fu2',
                            info='Over frequency shadding point 2',
                            unit='Hz',
                            )

        self.fu3 = NumParam(default=70,
                            tex_name='fu3',
                            info='Over frequency shadding point 3',
                            unit='Hz',
                            )

        self.Tfl1 = NumParam(default=300,
                             tex_name=r'T_{fl1}',
                             info='Stand time for (fl2, fl1)',
                             non_negative=True,
                             )

        self.Tfl2 = NumParam(default=10,
                             tex_name=r'T_{fl2}',
                             info='Stand time for (fl3, fl2)',
                             non_negative=True,
                             )

        self.Tfu1 = NumParam(default=300,
                             tex_name=r'T_{fu1}',
                             info='Stand time for (fu1, fu2)',
                             non_negative=True,
                             )

        self.Tfu2 = NumParam(default=10,
                             tex_name=r'T_{fu2}',
                             info='Stand time for (fu2, fu3)',
                             non_negative=True,
                             )

        # -- protection parameters, voltage
        self.vl4 = NumParam(default=0.1,
                            tex_name='vl4',
                            info='Under voltage shadding point 4',
                            unit='p.u.',
                            )

        self.vl3 = NumParam(default=0.45,
                            tex_name='vl3',
                            info='Under voltage shadding point 3',
                            unit='p.u.',
                            )

        self.vl2 = NumParam(default=0.6,
                            tex_name='vl2',
                            info='Under voltage shadding point 2',
                            unit='p.u.',
                            )

        self.vl1 = NumParam(default=0.88,
                            tex_name='vl1',
                            info='Under voltage shadding point 1',
                            unit='p.u.',
                            )

        self.vu1 = NumParam(default=1.1,
                            tex_name='vu1',
                            info='Over voltage shadding point 1',
                            unit='p.u.',
                            )

        self.vu2 = NumParam(default=1.2,
                            tex_name='vu2',
                            info='Over voltage shadding point 2',
                            unit='p.u.',
                            )

        self.vu3 = NumParam(default=2,
                            tex_name='vu3',
                            info='Over voltage shadding point 3',
                            unit='p.u.',
                            )

        self.Tvl1 = NumParam(default=2,
                             tex_name=r'T_{vl1}',
                             info='Stand time for (vl2, vl1)',
                             non_negative=True,
                             )

        self.Tvl2 = NumParam(default=1,
                             tex_name=r'T_{vl2}',
                             info='Stand time for (vl3, vl2)',
                             non_negative=True,
                             )

        self.Tvl3 = NumParam(default=0.16,
                             tex_name=r'T_{vl3}',
                             info='Stand time for (vl4, vl3)',
                             non_negative=True,
                             )

        self.Tvu1 = NumParam(default=1,
                             tex_name=r'T_{vu1}',
                             info='Stand time for (vu1, vu2)',
                             non_negative=True,
                             )

        self.Tvu2 = NumParam(default=0.16,
                             tex_name=r'T_{vu2}',
                             info='Stand time for (vu2, vu3)',
                             non_negative=True,
                             )

        self.Tres = NumParam(default=0.05,
                             info='Integrator reset time',
                             )


class DGPRCTBaseModel(Model):
    """
    Model implementation of DGPRCT Base.
    """

    def __init__(self, system, config):
        Model.__init__(self, system, config)
        self.flags.tds = True
        self.group = 'DGProtection'

        self.bus = ExtParam(model='DG', src='bus',
                            indexer=self.dev,
                            export=False)

        self.fn = ExtParam(model='DG', src='fn',
                           indexer=self.dev,
                           export=False)

        # -- Frequency protection
        # Convert frequency deviation range to p.u.
        self.f = ExtAlgeb(export=False,
                          info='DG frequency read value', unit='p.u.',
                          model='FreqMeasurement', src='f',
                          indexer=self.busfreq,
                          )

        self.fHz = Algeb(v_str='fn * f',
                         e_str='fn * f - fHz',
                         info='frequency in Hz',
                         tex_name=r'f_{Hz}',
                         )

        # -- Lock DG frequency signal and output power

        self.ltu = ConstService(v_str='0.8')
        self.ltl = ConstService(v_str='0.2')

        # `Ldsum_zu` is `ue`
        dsum = 'fen * (IAWfl1_lim_zu * Lfl1_zi + IAWfl2_lim_zu * Lfl2_zi + ' \
               'IAWfu1_lim_zu * Lfu1_zi + IAWfu2_lim_zu * Lfu2_zi) + ' \
               'Ven * (IAWVl1_lim_zu * LVl1_zi + IAWVl2_lim_zu * LVl2_zi + ' \
               'IAWVl3_lim_zu * LVl3_zi + ' \
               'IAWVu1_lim_zu * LVu1_zi + IAWVu2_lim_zu * LVu2_zi) - ' \
               'dsum'

        self.dsum = Algeb(v_str='0',
                          e_str=dsum,
                          info='lock signal summation',
                          tex_name=r'd_{tot}',
                          )

        self.Ldsum = Limiter(u=self.dsum,
                             lower=self.ltl,
                             upper=self.ltu,
                             info='lock signal comparer, zu is to act',
                             equal=False, no_warn=True,
                             )

        self.ue = Algeb(v_str='0',
                        e_str='Ldsum_zu - ue',
                        info='lock flag',
                        tex_name=r'ue',
                        )

        self.zero = ConstService('0')

        self.res = ExtendedEvent(self.ue, t_ext=self.Tres,
                                 trig="rise",
                                 extend_only=True)

        # lock DG frequency signal

        # fflag option 1: leave source signal online in protection
        self.fin = ExtAlgeb(model='DG', src='f',
                            indexer=self.dev,
                            info='original f from DG',
                            )

        self.fHzl = ExtAlgeb(model='DG', src='fHz',
                             indexer=self.dev,
                             export=False,
                             e_str='- ue * (fn * f)',
                             info='Frequency measure lock',
                             ename='fHzl',
                             tex_ename='f_{Hzl}',
                             )
        # TODO: add fflag option 2: block the source signal in protection

        # lock output power of DG

        self.Pext = ExtAlgeb(model='DG', src='Pext',
                             indexer=self.dev,
                             info='original Pext from DG',
                             )

        self.Pref = ExtAlgeb(model='DG', src='Pref',
                             indexer=self.dev,
                             info='original Pref from DG',
                             )

        self.Pdrp = ExtAlgeb(model='DG', src='DB_y',
                             indexer=self.dev,
                             info='original Pdrp from DG',
                             )

        self.Psum = ExtAlgeb(model='DG', src='Psum',
                             indexer=self.dev,
                             export=False,
                             e_str='- ue * (Pext + Pref + Pdrp)',
                             info='Active power lock',
                             ename='Pneg',
                             tex_ename='P_{neg}',
                             )

        self.Qdrp = ExtAlgeb(model='DG', src='Qdrp',
                             indexer=self.dev,
                             info='original Qdrp from DG',
                             )

        self.Qref = ExtAlgeb(model='DG', src='Qref',
                             indexer=self.dev,
                             info='original Qref from DG',
                             )

        self.Qsum = ExtAlgeb(model='DG', src='Qsum',
                             indexer=self.dev,
                             export=False,
                             e_str='- ue * (Qdrp + Qref)',
                             info='Reactive power lock',
                             ename='Qneg',
                             tex_ename='Q_{neg}',
                             )

        # TODO: apply ExtState here
        # TODO: clear State: BusFreq WO(Washout)
        # TODO: clear State: PVD1: Ipout, Iqout
        # TODO: develop integrator with reset function
        # TODO: Set EventFlag for ue


class fProtect:
    """
    Subclass for frequency protection logic.
    """

    def __init__(self):
        # Indicatior of frequency deviation
        self.Lfl1 = Limiter(u=self.fHz,
                            lower=self.fl3, upper=self.fl1,
                            info='Frequency comparer for (fl3, fl1)',
                            equal=False, no_warn=True,
                            )

        self.Lfl2 = Limiter(u=self.fHz,
                            lower=self.fl3, upper=self.fl2,
                            info='Frequency comparer for (fl3, fl2)',
                            equal=False, no_warn=True,
                            )

        self.Lfu1 = Limiter(u=self.fHz,
                            lower=self.fu1, upper=self.fu3,
                            info='Frequency comparer for (fu1, fu3)',
                            equal=False, no_warn=True,
                            )

        self.Lfu2 = Limiter(u=self.fHz,
                            lower=self.fu2, upper=self.fu3,
                            info='Frequency comparer for (fu2, fu3)',
                            equal=False, no_warn=True,
                            )

        # Frequency deviation time continuity check

        self.IAWfl1 = IntegratorAntiWindup(u='Lfl1_zi * (1 - res) - Tfl1 / Tres * res',
                                           T=1, K=1, y0='0',
                                           lower=self.zero, upper=self.Tfl1,
                                           info='condition check for (fl3, fl1)',
                                           no_warn=True,
                                           )

        self.IAWfl2 = IntegratorAntiWindup(u='Lfl2_zi * (1 - res) - Tfl2 / Tres * res',
                                           T=1, K=1, y0='0',
                                           lower=self.zero, upper=self.Tfl2,
                                           info='condition check for (fl3, fl2)',
                                           no_warn=True,
                                           )

        self.IAWfu1 = IntegratorAntiWindup(u='Lfu1_zi * (1 - res) - Tfu1 / Tres * res',
                                           T=1, K=1, y0='0',
                                           lower=self.zero, upper=self.Tfu1,
                                           info='condition check for (fu1, fu3)',
                                           no_warn=True,
                                           )

        self.IAWfu2 = IntegratorAntiWindup(u='Lfl2_zi * (1 - res) - Tfu2 / Tres * res',
                                           T=1, K=1, y0='0',
                                           lower=self.zero, upper=self.Tfu2,
                                           info='condition check for (fu2, fu3)',
                                           no_warn=True,
                                           )


class VProtect:
    """
    Subclass for voltage protection logic
    """

    def __init__(self):
        # -- Voltage protection

        # Indicatior of voltage deviation
        self.LVl1 = Limiter(u=self.v,
                            lower=self.vl4, upper=self.vl1,
                            info='Voltage comparer for (vl4, vl1)',
                            equal=False, no_warn=True,
                            )

        self.LVl2 = Limiter(u=self.v,
                            lower=self.vl4, upper=self.vl2,
                            info='Voltage comparer for (vl4, vl2)',
                            equal=False, no_warn=True,
                            )

        self.LVl3 = Limiter(u=self.v,
                            lower=self.vl4, upper=self.vl3,
                            info='Voltage comparer for (vl4, vl3)',
                            equal=False, no_warn=True,
                            )

        self.LVu1 = Limiter(u=self.v,
                            lower=self.vu1, upper=self.vu3,
                            info='Voltage comparer for (vu1, vu3)',
                            equal=False, no_warn=True,
                            )

        self.LVu2 = Limiter(u=self.v,
                            lower=self.vu2, upper=self.vu3,
                            info='Voltage comparer for (vu2, vu3)',
                            equal=False, no_warn=True,
                            )

        # Voltage deviation time continuity check
        self.IAWVl1 = IntegratorAntiWindup(u='LVl1_zi * (1 - res) - Tvl1 / Tres * res',
                                           T=1, K=1, y0='0',
                                           lower=self.zero, upper=self.Tvl1,
                                           info='condition check for (Vl3, Vl1)',
                                           no_warn=True,
                                           )

        self.IAWVl2 = IntegratorAntiWindup(u='LVl2_zi * (1 - res) - Tvl2 / Tres * res',
                                           T=1, K=1, y0='0',
                                           lower=self.zero, upper=self.Tvl2,
                                           info='condition check for (Vl3, Vl2)',
                                           no_warn=True,
                                           )

        self.IAWVl3 = IntegratorAntiWindup(u='LVl2_zi * (1 - res) - Tvl3 / Tres * res',
                                           T=1, K=1, y0='0',
                                           lower=self.zero, upper=self.Tvl2,
                                           info='condition check for (Vl3, Vl2)',
                                           no_warn=True,
                                           )

        self.IAWVu1 = IntegratorAntiWindup(u='LVu1_zi * (1 - res) - Tvu1 / Tres * res',
                                           T=1, K=1, y0='0',
                                           lower=self.zero, upper=self.Tvu1,
                                           info='condition check for (Vu1, Vu3)',
                                           no_warn=True,
                                           )

        self.IAWVu2 = IntegratorAntiWindup(u='LVu2_zi * (1 - res) - Tvu2 / Tres * res',
                                           T=1, K=1, y0='0',
                                           lower=self.zero, upper=self.Tvu2,
                                           info='condition check for (Vu2, Vu3)',
                                           no_warn=True,
                                           )


class DGPRCT1Model(DGPRCTBaseModel, fProtect, VProtect):
    """
    Model implementation of DGPRCT1.
    """

    def __init__(self, system, config):
        DGPRCTBaseModel.__init__(self, system, config)
        self.v = ExtAlgeb(model='Bus', src='v',
                          indexer=self.bus,
                          export=False,
                          info='Bus voltage',
                          unit='p.u.',
                          )
        fProtect.__init__(self)
        VProtect.__init__(self)


class DGPRCT1(DGPRCTBaseData, DGPRCT1Model):
    """
    DGPRCT1 model, follow IEEE-1547-2018. DGPRCT stands for DG protection.

    A demo is provided: examples/demonstration/1.1 demo_DGPRCT1.ipynb

    Target device (limited to DG group) ``Psum`` and ``Qsum`` will decrease to zero
    immediately when frequency/voltage protection flag is raised. Once the lock is
    released, ``Psum`` and ``Qsum`` will return to normal immediately.

    DG group base model ``PVD1`` already has a degrading function which is used to
    degrade output under abnormal condition. it is recommended to turn it off by setting
    `recflag = 0`.

    ``fen`` and ``Ven`` are protection enabling parameters. 1/0 is on/off.

    ``ue`` is lock flag signal.

    It should be noted that, the lock only lock the ``fHz`` (frequency read value)
    of DG model. The source values (which come from ``BusFreq`` `f` remain unchanged.)

    Protection sensors (e.g., IAWfl1) are instances of ``IntergratorAntiWindup``. All
    the protection sensors will be reset after ``ue`` returns to 0.
    Resetting action takes `Tres` to finish.

    The model does not check the shedding points sequence.
    The input parameters are required to satisfy ``fl3 < fl2 < fl1 < fu1 < fu2 < fu3``,
    and ``ul4 < ul3 < ul2 < ul1 < uu1 < uu2 < uu3``.

    Default settings:\n
    Frequency (Hz):\n
    `(fl3, fl2), Tfl2` [(50.0, 57.5), 10s]\n
    `(fl2, fl1), Tfl1` [(57.5, 59.2), 300s]\n
    `(fu1, fu2), Tfu1` [(60.5, 61.5), 300s]\n
    `(fu2, fu3), Tfu2` [(61.5, 70.0), 10s]\n

    Voltage (p.u.):\n
    `(vl4, vl3), Tvl3` [(0.10, 0.45), 0.16s]\n
    `(vl3, vl2), Tvl2` [(0.45, 0.60), 1s]\n
    `(vl2, vl1), Tvl1` [(0.60, 0.88), 2s]\n
    `(vu1, vu2), Tvu1` [(1.10, 1.20), 1s]\n
    `(vu2, vu3), Tvu2` [(1.20, 2.00), 0.16s]\n

    Reference:

    NERC. Bulk Power System Reliability Perspectives on the Adoption of IEEE 1547-2018.
    March 2020. Available:

    https://www.nerc.com/comm/PC_Reliability_Guidelines_DL/Guideline_IEEE_1547-2018_BPS_Perspectives.pdf
    """

    def __init__(self, system, config):
        DGPRCTBaseData.__init__(self)
        DGPRCT1Model.__init__(self, system, config)


class DGPRCTExtModel(DGPRCTBaseModel, fProtect, VProtect):
    """
    Model implementation of DGPRCT1.
    """

    def __init__(self, system, config):
        DGPRCTBaseModel.__init__(self, system, config)

        self.v = ExtService(model='Bus', src='v',
                            indexer=self.bus,
                            info='retrived voltage',
                            tex_name='v',
                            )

        fProtect.__init__(self)
        VProtect.__init__(self)


class DGPRCTExt(DGPRCTBaseData, DGPRCTExtModel):
    """
    DGPRCT External model, follow IEEE-1547-2018. DGPRCT stands for DG protection.

    Similar to DGPRCT1, but the measured voltage can be manipulated.

    A demo is provided: examples/demonstration/1.2 demo_DGPRCTExt.ipynb

    This model can be applied to co-simulation, where you can input the external votage signal
    into ANDES. If no extertal value is applied, the votalge will remain as the initialized value.

    Target device (limited to DG group) ``Psum`` and ``Qsum`` will decrease to zero
    immediately when frequency/voltage protection flag is raised. Once the lock is
    released, ``Psum`` and ``Qsum`` will return to normal immediately.

    DG group base model ``PVD1`` already has a degrading function which is used to
    degrade output under abnormal condition. it is recommended to turn it off by setting
    `recflag = 0`.

    ``fen`` and ``Ven`` are protection enabling parameters. 1/0 is on/off.

    ``ue`` is lock flag signal.

    It should be noted that, the lock only lock the ``fHz`` (frequency read value)
    of DG model. The source values (which come from ``BusFreq`` `f` remain unchanged.)

    Protection sensors (e.g., IAWfl1) are instances of ``IntergratorAntiWindup``. All
    the protection sensors will be reset after ``ue`` returns to 0.
    Resetting action takes `Tres` to finish.

    The model does not check the shedding points sequence.
    The input parameters are required to satisfy ``fl3 < fl2 < fl1 < fu1 < fu2 < fu3``,
    and ``ul4 < ul3 < ul2 < ul1 < uu1 < uu2 < uu3``.

    Default settings:\n
    Frequency (Hz):\n
    `(fl3, fl2), Tfl2` [(50.0, 57.5), 10s]\n
    `(fl2, fl1), Tfl1` [(57.5, 59.2), 300s]\n
    `(fu1, fu2), Tfu1` [(60.5, 61.5), 300s]\n
    `(fu2, fu3), Tfu2` [(61.5, 70.0), 10s]\n

    Voltage (p.u.):\n
    `(vl4, vl3), Tvl3` [(0.10, 0.45), 0.16s]\n
    `(vl3, vl2), Tvl2` [(0.45, 0.60), 1s]\n
    `(vl2, vl1), Tvl1` [(0.60, 0.88), 2s]\n
    `(vu1, vu2), Tvu1` [(1.10, 1.20), 1s]\n
    `(vu2, vu3), Tvu2` [(1.20, 2.00), 0.16s]\n

    Reference:

    NERC. Bulk Power System Reliability Perspectives on the Adoption of IEEE 1547-2018.
    March 2020. Available:

    https://www.nerc.com/comm/PC_Reliability_Guidelines_DL/Guideline_IEEE_1547-2018_BPS_Perspectives.pdf
    """

    def __init__(self, system, config):
        DGPRCTBaseData.__init__(self)
        DGPRCTExtModel.__init__(self, system, config)



import io
import logging
import os

logger = logging.getLogger(__name__)


class FileMan:
    """
    Define a File Manager class for System
    """

    def __init__(self, case=None, **kwargs):
        """
        Initialize the output file names.
        For inputs, all absolute paths will be respected.
        All relative paths are relative to `input_path`.

        case: must be full path to case

        output: desired name for format conversion output

        input_path: default path for input files that only contains file name. If `input_path` is not provided,
                    it will be derived from the path of `case`.

        output_path: path for output files. Default to current working directory where `andes` is invoked.
        """
        self.input_format = None
        self.output_format = None
        self.add_format = None

        self.case = None
        self.input_path = ''
        self.case_path = ''  # absolute path containinng the case file
        self.fullname = None
        self.name = None
        self.ext = None
        self.addfile = None
        self.pert = None
        self.output_path = None
        self.no_output = True

        self._out_fields = ['txt', 'dump', 'lst', 'eig', 'npy', 'npz', 'csv', 'mat', 'prof', 'prof_raw']
        self.set(case, **kwargs)

    def set(self, case=None, **kwargs):
        """
        Perform the input and output set up.
        """
        self.case = case
        input_format = kwargs.get('input_format')
        add_format = kwargs.get('add_format')
        input_path = kwargs.get('input_path')
        addfile = kwargs.get('addfile')
        no_output = kwargs.get('no_output')
        output_path = kwargs.get('output_path')
        output = kwargs.get('output')  # base file name for the output
        pert = kwargs.get('pert')
        dump = kwargs.get('dump')

        # set output fields anyway
        for item in self._out_fields:
            self.__dict__[item] = None

        if case is None:
            return

        self.input_format = input_format
        self.add_format = add_format
        self.input_path = input_path if input_path is not None else ''
        self.output_path = output_path if output_path is not None else ''

        self.case = case
        if os.path.isfile(case):
            self.case_path, self.fullname = os.path.split(self.case)
            self.case_path = os.path.abspath(self.case_path)
        else:
            self.fullname = ''

        # `self.name` is the name part without extension
        self.name, self.ext = os.path.splitext(self.fullname)

        self.addfile = self.get_fullpath(addfile)
        self.pert = self.get_fullpath(pert)
        if dump is None:
            dump = self.name

        # set the output file names anyway
        if not output:
            output = add_suffix(self.name, 'out')
        prof = add_suffix(self.name, 'prof')
        eig = add_suffix(self.name, 'eig')
        mat = add_suffix(self.name, 'As')

        self.lst = os.path.join(self.output_path, output + '.lst')
        self.npy = os.path.join(self.output_path, output + '.npy')
        self.npz = os.path.join(self.output_path, output + '.npz')
        self.csv = os.path.join(self.output_path, output + '.csv')
        self.txt = os.path.join(self.output_path, output + '.txt')

        self.eig = os.path.join(self.output_path, eig + '.txt')
        self.prof = os.path.join(self.output_path, prof + '.txt')
        self.mat = os.path.join(self.output_path, mat + '.mat')
        self.prof_raw = os.path.join(self.output_path, prof + '.prof')
        self.dump = os.path.join(self.output_path, dump + '.xlsx')

        # use `self.no_output` to control calls to saving functions
        if no_output:
            self.no_output = True
        else:
            self.no_output = False

    def get_fullpath(self, fullname=None):
        """
        Return the original full path if full path is specified, otherwise
        search in the case file path.

        Parameters
        ----------
        fullname : str, optional
            Full name of the file. If relative, prepend `input_path`.
            Otherwise, leave it as is.
        """
        # if is an empty path
        if not fullname:
            return fullname
        if isinstance(fullname, io.IOBase):
            return ''

        isabs = os.path.isabs(fullname)

        path, name = os.path.split(fullname)

        if not name:  # path to a folder
            return ''
        else:  # path to a file
            if isabs:
                return fullname
            else:
                return os.path.join(self.input_path, path, name)


def add_suffix(fullname, suffix):
    """
    Add suffix to a full file name.
    """

    name, ext = os.path.splitext(fullname)
    return name + '_' + suffix + ext



"""
Integration methods for DAE.
"""

import logging
import numpy as np

from andes.shared import sparse, matrix, tqdm


logger = logging.getLogger(__name__)


class ImplicitIter:
    """
    Base class for implicit iterative methods.
    """

    @staticmethod
    def calc_jac(tds, gxs, gys):
        pass

    @staticmethod
    def calc_q(x, f, Tf, h, x0, f0):
        pass

    @staticmethod
    def step(tds):
        """
        Integrate with Implicit Trapezoidal Method (ITM) to the current time.

        This function has an internal Newton-Raphson loop for algebraized semi-explicit DAE.
        The function returns the convergence status when done but does NOT progress simulation time.

        Returns
        -------
        bool
            Convergence status in ``tds.converged``.

        """
        system = tds.system
        dae = tds.system.dae

        if tds.h == 0:
            logger.error("Current step size is zero. Integration is not permitted.")
            return False

        tds.mis = [1, 1]
        tds.niter = 0
        tds.converged = False

        tds.x0[:] = dae.x
        tds.y0[:] = dae.y
        tds.f0[:] = dae.f

        while True:
            tds.fg_update(models=system.exist.pflow_tds)

            # lazy Jacobian update

            reason = ''
            if dae.t == 0:
                reason = 't=0'
            elif tds.config.honest:
                reason = 'using honest method'
            elif tds.custom_event:
                reason = 'custom event set'
            elif not tds.last_converged:
                reason = 'non-convergence in the last step'
            elif tds.niter > 4 and (tds.niter + 1) % 3 == 0:
                reason = 'every 3 iterations beyond 4 iterations'
            elif dae.t - tds._last_switch_t < 0.1:
                reason = 'within 0.1s of event'

            if reason:
                system.j_update(models=system.exist.pflow_tds, info=reason)

                # set flag in `solver.worker.factorize`, not `solver.factorize`.
                tds.solver.worker.factorize = True

            # `Tf` should remain constant throughout the simulation, even if the corresponding diff. var.
            # is pegged by the anti-windup limiters.

            # solve implicit trapezoidal method (ITM) integration
            if tds.config.g_scale > 0:
                gxs = tds.config.g_scale * tds.h * dae.gx
                gys = tds.config.g_scale * tds.h * dae.gy
            else:
                gxs = dae.gx
                gys = dae.gy

            # calculate complete Jacobian matrix ``Ac```
            tds.Ac = tds.method.calc_jac(tds, gxs, gys)

            # equation `tds.qg[:dae.n] = 0` is the implicit form of differential equations using ITM
            tds.qg[:dae.n] = tds.method.calc_q(dae.x, dae.f, dae.Tf, tds.h, tds.x0, tds.f0)

            # reset the corresponding q elements for pegged anti-windup limiter
            for item in system.antiwindups:
                for key, _, eqval in item.x_set:
                    np.put(tds.qg, key, eqval)

            # set or scale the algebraic residuals
            if tds.config.g_scale > 0:
                tds.qg[dae.n:] = tds.config.g_scale * tds.h * dae.g
            else:
                tds.qg[dae.n:] = dae.g

            # calculate variable corrections
            if not tds.config.linsolve:
                inc = tds.solver.solve(tds.Ac, matrix(tds.qg))
            else:
                inc = tds.solver.linsolve(tds.Ac, matrix(tds.qg))

            # check for np.nan first
            if np.isnan(inc).any():
                tds.err_msg = 'NaN found in solution. Convergence is not likely'
                tds.niter = tds.config.max_iter + 1
                tds.busted = True
                break

            # reset tiny values to reduce chattering
            if tds.config.reset_tiny:
                inc[np.where(np.abs(inc) < tds.tol_zero)] = 0

            # set new values
            dae.x -= inc[:dae.n].ravel()
            dae.y -= inc[dae.n: dae.n + dae.m].ravel()

            # synchronize solutions to model internal storage
            system.vars_to_models()

            # store `inc` to tds for debugging
            tds.inc = inc

            mis = np.max(np.abs(inc))

            # store initial maximum mismatch
            if tds.niter == 0:
                tds.mis[0] = mis
            else:
                tds.mis[-1] = mis

            tds.niter += 1

            # converged
            if mis <= tds.config.tol:
                tds.converged = True
                break

            # non-convergence cases
            if tds.niter > tds.config.max_iter:
                if system.options.get("verbose", 20) <= 10:
                    tqdm.write(f'* Max. iter. {tds.config.max_iter} reached for t={dae.t:.6f}s, '
                               f'h={tds.h:.6f}s, max inc={mis:.4g} ')

                # debug helpers
                g_max = np.argmax(abs(dae.g))
                inc_max = np.argmax(abs(inc))
                tds._debug_g(g_max)
                tds._debug_ac(inc_max)
                break

            if (mis > 1e6) and (mis > 1e6 * tds.mis[0]):
                tds.err_msg = 'Error increased too quickly.'
                break

        if not tds.converged:
            dae.x[:] = np.array(tds.x0)
            dae.y[:] = np.array(tds.y0)
            dae.f[:] = np.array(tds.f0)
            system.vars_to_models()

        tds.last_converged = tds.converged

        return tds.converged


class BackEuler(ImplicitIter):
    """
    Backward Euler's integration method.
    """
    @staticmethod
    def calc_jac(tds, gxs, gys):
        """
        Build full Jacobian matrix ``Ac`` for Trapezoid method.
        """

        dae = tds.system.dae

        return sparse([[tds.Teye - tds.h * dae.fx, gxs],
                       [-tds.h * dae.fy, gys]], 'd')

    @staticmethod
    def calc_q(x, f, Tf, h, x0, f0):
        """
        Calculate the residual of algebraized differential equations.

        Notes
        -----
        Numba jit somehow slows down this function for the 14-bus
        and the 2k-bus systems.
        """

        return Tf * (x - x0) - h * f


class Trapezoid(ImplicitIter):
    """
    Trapezoidal methods.
    """

    @staticmethod
    def calc_jac(tds, gxs, gys):
        """
        Build full Jacobian matrix ``Ac`` for Trapezoid method.
        """

        dae = tds.system.dae

        return sparse([[tds.Teye - tds.h * 0.5 * dae.fx, gxs],
                       [-tds.h * 0.5 * dae.fy, gys]], 'd')

    @staticmethod
    def calc_q(x, f, Tf, h, x0, f0):
        """
        Calculate the residual of algebraized differential equations.

        Notes
        -----
        Numba jit somehow slows down this function for the 14-bus
        and the 2k-bus systems.
        """

        return Tf * (x - x0) - h * 0.5 * (f + f0)


# --- solution method name-to-class mapping ---
# !!! add new solvers to below

method_map = {"trapezoid": Trapezoid,
              "backeuler": BackEuler,
              }



#  [ANDES] (C)2015-2022 Hantao Cui
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  File name: service.py
#  Last modified: 8/16/20, 7:28 PM

import logging
from collections import OrderedDict
from typing import Callable, Optional, Type, Union

import numpy as np

from andes.core.common import dummify
from andes.core.param import BaseParam
from andes.utils.func import list_flatten
from andes.utils.tab import Tab

logger = logging.getLogger(__name__)


class BaseService:
    """
    Base class for Service.

    Service is a v-provider type for holding internal and temporary values.
    Subclasses need to implement ``v``
    as a member attribute or using a property decorator.

    Parameters
    ----------
    name : str
        Instance name

    Attributes
    ----------
    owner : Model
        The hosting/owner model instance
    """

    def __init__(self, name: str = None, tex_name: str = None, unit: str = None,
                 info: str = None, vtype: Type = None):
        self.name = name
        self.tex_name = tex_name if tex_name else name
        self.info = info
        self.unit = unit
        self.vtype = vtype if vtype is not None else float  # type for `v`
        self.owner = None

    def get_names(self):
        """
        Return `name` in a list

        Returns
        -------
        list
            A list only containing the name of the service variable
        """
        return [self.name]

    def assign_memory(self, n):
        """
        Assign memory for ``self.v`` and set the array to zero.

        Parameters
        ----------
        n : int
            Number of elements of the value array.
            Provided by caller (Model.list2array).
        """
        self.v = np.zeros(n, dtype=self.vtype)

    @property
    def class_name(self):
        """
        Return the class name
        """
        return self.__class__.__name__

    @property
    def n(self):
        """
        Return the count of values in ``self.v``.

        Needs to be overloaded if ``v`` of subclasses is not a 1-dimensional
        array.

        Returns
        -------
        int
            The count of elements in this variable
        """
        if isinstance(self.v, (list, np.ndarray)):
            return len(self.v)
        else:
            return 1

    def __repr__(self):
        val_str = ''

        if (1 <= self.n <= 20) and hasattr(self, 'v'):
            val_str = f', v={self.v}'

        return f'{self.class_name}: {self.owner.class_name}.{self.name}{val_str}'


class ConstService(BaseService):
    """
    A type of Service that stays constant once initialized.

    ConstService are usually constants calculated from parameters. They are only
    evaluated once in the initialization phase before variables are initialized.
    Therefore, uninitialized variables must not be used in `v_str``.

    ConstService are evaluated *in sequence* after getting external variables
    and parameters and before initializing internal variables.

    Parameters
    ----------
    name : str
        Name of the ConstService
    v_str : str
        An equation string to calculate the variable value.
    v_numeric : Callable, optional
        A callable which returns the value of the ConstService
    v_type: type, optional, default to float
        Type of element in the value array in float or complex

    Attributes
    ----------
    v : array-like or a scalar
        ConstService value
    """

    def __init__(self,
                 v_str: Optional[str] = None,
                 v_numeric: Optional[Callable] = None,
                 vtype: Optional[type] = None,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 ):
        super().__init__(name=name, vtype=vtype, tex_name=tex_name, info=info,
                         unit=unit)
        self.v_str: str = v_str
        self.v_numeric: Callable = v_numeric
        self.v: Union[float, int, np.ndarray] = 0.0
        self.sequential = True


class SubsService(BaseService):
    """
    A service to be substituted by its ``v_str`` in equations where it appears.

    ``SubsService`` is useful for eliminating explicit algebraic variables from
    equations.

    Examples
    --------

    If one defines the following inside ``__init__()``:

    .. code:: python

        self.vd = SubsService(v_str='v * cos(delta - a)') self.p =
        Algeb(e_str='vd * Id + vq * Iq')

    At code-generation time, in ``p`` 's equation, the ``vd`` variable will be
    replaced by its full expression, namely, ``v * cos(delta - a)``.
    """

    def __init__(self,
                 v_str: Optional[str] = None,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 ):
        super().__init__(name=name, tex_name=tex_name, info=info, unit=unit)
        self.v_str: str = v_str


class VarService(ConstService):
    """
    Variable service that gets updated in each step/iteration before computing
    the residual equations. As a results, variable values from the k-th step are
    used to compute a ``VarService`` that will be used to compute the residual
    for the (k+1)-th step.

    This class is useful when one has non-differentiable algebraic equations,
    which make use of `abs()`, `re` and `im`. Instead of creating `Algeb`, one
    can put the equation in `VarService`, which will be updated before solving
    algebraic equations.

    Parameters
    ----------
    sequential : bool, optional, default to True
        True if this VarService depends on previously defined VarService and
        should be evaluated in sequence. False if this VarService only uses known
        variables.

    Examples
    --------
    In ESST3A model, the voltage and current sensors (vd + jvq), (Id + jIq)
    estimate the sensed VE using equation

    .. math ::

        VE = | K_{PC}*(v_d + 1j v_q) + 1j (K_I + K_{PC}*X_L)*(I_d + 1j I_q)|

    One can use `VarService` to implement this equation ::

        self.VE = VarService(
            tex_name='V_E', info='VE', v_str='Abs(KPC*(vd + 1j*vq) + 1j*(KI +
            KPC*XL)*(Id + 1j*Iq))', )

    Warnings
    --------
    `VarService` is not solved with other algebraic equations, meaning that
    there is one step "delay" between the algebraic variables and `VarService`.
    Use an algebraic variable whenever possible.
    """

    def __init__(self,
                 v_str: Optional[str] = None,
                 v_numeric: Optional[Callable] = None,
                 vtype: Optional[type] = None,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 sequential: Optional[bool] = True,
                 ):

        super().__init__(name=name,
                         vtype=vtype,
                         tex_name=tex_name,
                         info=info,
                         unit=unit,
                         v_str=v_str,
                         v_numeric=v_numeric)
        self.sequential = sequential


class EventFlag(VarService):
    """
    Service to flag events when the input value changes. The typical input is a
    `v-provider` with binary values.

    Implemented by providing `self.check(**kwargs)` as `v_numeric`.
    `EventFlag.v` stores the values of the input variable in the most recent
    iteration/step.

    After the evaluation of `self.check()`, `self.v` will be updated.
    """

    def __init__(self, u,
                 vtype: Optional[type] = None,
                 name: Optional[str] = None, tex_name=None, info=None):
        VarService.__init__(self, v_numeric=self.check,
                            vtype=vtype, name=name, tex_name=tex_name, info=info)
        self.u = dummify(u)

    def check(self, **kwargs):
        """
        Check status and set event flags.

        Input values are compared with values in the memory.
        """
        if not np.all(self.v == self.u.v):
            self.owner.system.TDS.custom_event = True
            logger.debug(f"Event flag set at t={self.owner.system.dae.t:.6f} sec.")

        return self.u.v


class VarHold(VarService):
    """
    Service for holding the input when the hold signal is on.

    Parameters
    ----------
    hold : v-provider, binary
        Hold signal array with length equal to the input. For elements that are
        1, the corresponding inputs are held until the hold signal returns to 0.
    """

    def __init__(self, u, hold, vtype=None, name=None, tex_name=None, info=None):
        VarService.__init__(self, v_numeric=self.check, vtype=vtype,
                            name=name, tex_name=tex_name, info=info,
                            )
        self.u = dummify(u)
        self.hold = dummify(hold)
        self._init = False

    def check(self, **kwargs):
        """
        Custom `v_numeric` function for checking the
        hold signal and calculating outputs.
        """
        if not np.all(self.hold.v == 0.0):
            hold_idx = np.where(self.hold.v == 1)

            ret = self.u.v.copy()
            ret[hold_idx] = self.v[hold_idx]

            return ret

        else:
            return self.u.v


class ExtendedEvent(VarService):
    """
    Service for indicating an event for an extended, predefined period of time
    following the event disappearance.

    The triggering of an event, whether the rise or fall edge, is specified
    through `trig`. For example, if `trig = rise`, the change of the input from
    0 to 1 will be considered as an input, whereas the subsequent change back to
    0 will be considered as the event end.

    `ExtendedEvent.v` stores the flags whether the extended time has completed.
    Outputs will become 1 once the event starts and return to 0 when the
    extended time ends.

    Warnings
    --------
    The performance of this class needs to be optimized.

    Parameters
    ----------
    u : v-provider
        Triggering signal where the values are 0 or 1.

    trig : str in ("rise", "fall")
        Triggering edge for the beginning of an event. `rise` by default.

    enable : bool or v-provider
        If disabled, the output will be `v_disabled`

    extend_only : bool
        Only output during the extended period, not the event period.
    """

    def __init__(self,
                 u,
                 t_ext: Union[int, float, BaseParam, BaseService] = 0.0,
                 trig: str = 'rise',
                 enable=True,
                 v_disabled=0,
                 extend_only=False,
                 vtype: Optional[type] = None,
                 name: Optional[str] = None, tex_name=None, info=None):
        VarService.__init__(self, v_numeric=self.check,
                            vtype=vtype, name=name, tex_name=tex_name, info=info)

        self.u = dummify(u)
        self.t_ext = dummify(t_ext)
        self.enable = dummify(enable)
        self.v_disabled = v_disabled
        self.extend_only = extend_only

        self.t_final = None
        self.trig = trig

        self.v_event = None
        self.u_last = None
        self.z = None  # if is in an extended event (from event start to extension end)
        self.n_ext = 0  # number of extended events

    def assign_memory(self, n):
        """
        Assign memory for internal data.
        """
        VarService.assign_memory(self, n)
        self.t_final = np.zeros_like(self.v)
        self.v_event = np.zeros_like(self.v)
        self.u_last = np.zeros_like(self.v)
        self.z = np.zeros_like(self.v)

        if isinstance(self.t_ext.v, (int, float)):
            self.t_ext.v = np.ones_like(self.u.v) * self.t_ext.v

    def check(self, **kwargs):
        """
        Check if an extended event is in place.

        Supplied as a ``v_numeric`` to ``VarService``.
        """
        dae_t = self.owner.system.dae.t

        if dae_t == 0.0:
            self.u_last[:] = self.u.v
            self.v_event[:] = self.u.v

        # when any input signal changes
        if not np.all(self.u.v == self.u_last):
            diff = self.u.v - self.u_last

            # detect the actual ending of an event
            if self.trig == 'rise':
                starting = np.where(diff == 1)[0]
                ending = np.where(diff == -1)[0]
            else:
                starting = np.where(diff == -1)[0]
                ending = np.where(diff == 1)[0]

            if len(starting):
                self.z[starting] = 1

                if not self.extend_only:
                    self.v_event[starting] = self.u.v[starting]

            if len(ending):
                if self.extend_only:
                    self.v_event[ending] = self.u_last[ending]

                final_times = dae_t + self.t_ext.v[ending]
                self.t_final[ending] = final_times

                self.n_ext += len(ending)

                # TODO: insert extended event end times to a model-level list
                logger.debug("Extended Event ending time set at t=%s sec.",
                             str(final_times))

        # final time of the extended event
        if self.n_ext and np.any(self.t_final <= dae_t):
            self.z[np.where(self.t_final <= dae_t)] = 0
            self.n_ext = np.count_nonzero(self.z)

        self.u_last[:] = self.u.v

        return self.enable.v * (self.u.v * (1 - self.z) + self.v_event * self.z) + \
            (1-self.enable.v) * self.v_disabled


class PostInitService(ConstService):
    """
    Constant service that gets stored once after init.

    This service is useful when one need to store initialization values stored
    in variables.

    Examples
    --------
    In ESST3A model, the `vf` variable is initialized followed by other
    variables. One can store the initial `vf` into `vf0` so that equation ``vf -
    vf0 = 0`` will hold. ::

        self.vref0 = PostInitService(info='Initial reference voltage input',
                                     tex_name='V_{ref0}', v_str='vref', )

    Since all `ConstService` are evaluated before equation evaluation, without
    using PostInitService, one will need to create lots of `ConstService` to
    store values in the initialization path towards `vf0`, in order to correctly
    initialize `vf`.
    """

    pass


class BackRef(BaseService):
    """
    A special type of reference collector.

    `BackRef` is used for collecting device indices of other models referencing
    the parent model of the `BackRef`. The `v``field will be a list of lists,
    each containing the `idx` of other models referencing each device of the
    parent model.

    BackRef can be passed as indexer for params and vars, or shape for
    `NumReduce` and `NumRepeat`. See examples for illustration.

    Examples
    --------
    A Bus device has an `IdxParam` of `area`, storing the `idx` of area to which
    the bus device belongs. In ``Bus.__init__()``, one has ::

        self.area = IdxParam(model='Area')

    Suppose `Bus` has the following data

        ====   ====  ====
        idx    area  Vn
        ----   ----  ----
        1      1     110
        2      2     220
        3      1     345
        4      1     500
        ====   ====  ====

    The Area model wants to collect the indices of Bus devices which points to
    the corresponding Area device. In ``Area.__init__``, one defines ::

        self.Bus = BackRef()

    where the member attribute name `Bus` needs to match exactly model name that
    `Area` wants to collect `idx` for. Similarly, one can define
    ``self.ACTopology = BackRef()`` to collect devices in the `ACTopology` group
    that references Area.

    The collection of `idx` happens in
    :py:func:`andes.system.System._collect_ref_param`. It has to be noted that
    the specific `Area` entry must exist to collect model idx-dx referencing it.
    For example, if `Area` has the following data ::

        idx 1

    Then, only Bus 1, 3, and 4 will be collected into `self.Bus.v`, namely,
    ``self.Bus.v == [ [1, 3, 4] ]``.

    If `Area` has data ::

        idx 1 2

    Then, `self.Bus.v` will end up with ``[ [1, 3, 4], [2] ]``.

    See Also
    --------
    andes.core.service.NumReduce : A more complete example using BackRef to
        build the COI model

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.export = False
        self.v = list()


class ExtService(BaseService):
    """
    Service constants whose value is from an external model or group.

    Parameters
    ----------
    src : str
        Variable or parameter name in the source model or group
    model : str
        A model name or a group name
    indexer : IdxParam or BaseParam
        An "Indexer" instance whose ``v`` field contains the ``idx`` of devices
        in the model or group.

    Examples
    --------
    A synchronous generator needs to retrieve the ``p`` and ``q`` values from
    static generators for initialization. ``ExtService`` is used for this
    purpose.

    In a synchronous generator, one can define the following to retrieve
    ``StaticGen.p`` as ``p0``::

        class GENCLSModel(Model):
            def __init__(...):
                ...
                self.p0 = ExtService(src='p',
                                     model='StaticGen',
                                     indexer=self.gen,
                                     tex_name='P_0')

    """

    def __init__(self,
                 model: str,
                 src: str,
                 indexer: Union[BaseParam, BaseService],
                 attr: str = 'v',
                 allow_none: bool = False,
                 default=0,
                 name: str = None,
                 tex_name: str = None,
                 vtype=None,
                 info: str = None,
                 ):
        super().__init__(name=name, tex_name=tex_name, info=info, vtype=vtype)
        self.model = model
        self.src = src
        self.indexer = indexer
        self.attr = attr
        self.allow_none = allow_none
        self.default = default
        self.v = np.array([0.])

    def link_external(self, ext_model):
        """
        Method to be called by ``System`` for getting values from the external
        model or group.

        Parameters
        ----------
        ext_model
            An instance of a model or group provided by System
        """
        # set initial v values to zero
        self.v = np.zeros(self.n)
        if self.n == 0:
            return

        # the same `get` api for Group and Model
        self.v = ext_model.get(src=self.src, idx=self.indexer.v, attr=self.attr,
                               allow_none=self.allow_none,
                               default=self.default,
                               )


class DataSelect(BaseService):
    """
    Class for selecting values for optional DataParam or NumParam.

    This service is a v-provider that uses optional DataParam when available.
    Otherwise, use the fallback value.

    DataParam will be tested for `None`, and NumParam will be tested with
    ``np.isnan()``.

    Notes
    -----
    An use case of DataSelect is remote bus. One can do

    .. code:: python

        self.buss = DataSelect(option=self.busr, fallback=self.bus)

    Then, pass ``self.buss`` instead of ``self.bus`` as indexer to retrieve
    voltages.

    Another use case is to allow an optional turbine rating. One can do

    .. code:: python

        self.Tn = NumParam(default=None)
        self.Sg = ExtParam(...)
        self.Sn = DataSelect(Tn, Sg)

    """

    def __init__(self,
                 optional,
                 fallback,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 ):
        super().__init__(name=name, tex_name=tex_name, info=info, )
        self.optional = optional
        self.fallback = fallback
        self._v = None

    @property
    def v(self):
        if self._v is None:
            self._v = [v1 if v1 is not None and not np.isnan(v1)
                       else v2
                       for v1, v2 in zip(self.optional.v, self.fallback.v)]

        return self._v


class DeviceFinder(BaseService):
    """
    Service for finding indices of optionally linked devices.

    If not provided, `DeviceFinder` will add devices at the beginning of
    `System.setup`.

    Examples
    --------
    IEEEST stabilizer takes an optional `busf` (IdxParam) for specifying the
    connected BusFreq, which is needed for mode 6. To avoid reimplementing
    `BusFreq` within IEEEST, one can do

    .. code-block :: python

        self.busfreq = DeviceFinder(self.busf,
                                    link=self.buss, idx_name='bus',
                                    default_model='BusFreq')

    where ``self.busf`` is the optional input, ``self.buss`` is the bus indices
    that ``busf`` should measure, and ``idx_name`` is the name of a BusFreq
    parameter through which the measured bus indices are specified. For each
    ``None`` values in ``self.busf``, a ``BusFreq`` device will be created to
    measure the corresponding bus in ``self.buss``.

    That is, ``BusFreq.[idx_name].v = [link]``. ``DeviceFinder`` will find or
    create ``BusFreq`` devices so that the returned list of ``BusFreq`` indices
    are connected to `self.buss`, respectively.
    """

    def __init__(self, u, link, idx_name, default_model,
                 name=None, tex_name=None, info=None):
        super().__init__(name=name, tex_name=tex_name, info=info)

        self.u = u
        self.model = u.model
        self.idx_name = idx_name
        self.default_model = default_model

        if self.model is None:
            raise ValueError(f'{u.owner.class_name}.{u.name} must contain "model".')

        self.link = link

    def find_or_add(self, system):
        """
        Find or add devices.

        Points `self.u.v` to the found or newly added devices.

        Find devices one by one.
        Devices previously added in this function can be used later without duplication.
        """

        if self.model in system.models:
            is_model = True
        elif self.model in system.groups:
            is_model = False
        else:
            raise ValueError('<%s> is not a valid model or group name.' % self.model)

        add_to_model = self.model if is_model else self.default_model

        mdl = system.__dict__[self.model]
        added = False

        for ii, link_to in enumerate(self.link.v):
            idx = mdl.find_idx(self.idx_name, (link_to, ), allow_none=True, default=None)[0]

            if idx is None:
                new_idx = system.add(add_to_model, {self.idx_name: link_to})
                self.u.v[ii] = new_idx

                logger.info(f"{self.owner.class_name} <{self.owner.idx.v[ii]}> "
                            f"added {add_to_model} <{new_idx}> "
                            f"linked to {self.idx_name} <{link_to}>")
                added = True
            else:
                self.u.v[ii] = idx

        if added:
            mdl = system.models[add_to_model]
            mdl.list2array()
            mdl.refresh_inputs()
            system.link_ext_param({mdl.name: mdl})

    @property
    def v(self):
        return self.u.v


class OperationService(BaseService):
    """
    Base class for a type of Service which performs specific operations.
    OperationService may not use the `assign_memory` from `BaseService`, because
    it can have a different size.

    This class cannot be used by itself.

    See Also
    --------
    NumReduce : Service for Reducing linearly stored 2-D services into 1-D

    NumRepeat : Service for repeating 1-D NumParam/ v-array following a
    sub-pattern

    IdxRepeat : Service for repeating 1-D IdxParam/ v-list following a
    sub-pattern
    """

    def __init__(self,
                 name=None,
                 tex_name=None,
                 info=None,
                 ):
        self._v = None
        super().__init__(name=name, tex_name=tex_name, info=info, )
        self.v_str = None

    @property
    def v(self):
        """
        Return values stored in `self._v`. May be overloaded by subclasses.
        """
        return self._v

    @v.setter
    def v(self, value):
        self._v = value


class NumReduce(OperationService):
    """
    A helper Service type which reduces a linearly stored 2-D ExtParam into 1-D
    Service.

    NumReduce works with ExtParam whose `v` field is a list of lists. A reduce
    function which takes an array-like and returns a scalar need to be supplied.
    NumReduce calls the reduce function on each of the lists and return all the
    scalars in an array.

    Parameters
    ----------
    u : ExtParam
        Input ExtParam whose ``v`` contains linearly stored 2-dimensional values
    ref : BackRef
        The BackRef whose 2-dimensional shapes are used for indexing
    fun : Callable
        The callable for converting a 1-D array-like to a scalar

    Examples
    --------
    Suppose one wants to calculate the mean value of the ``Vn`` in one Area. In
    the ``Area`` class, one defines ::

        class AreaModel(...):
            def __init__(...):
                ...
                # backward reference from `Bus`
                self.Bus = BackRef()

                # collect the Vn in an 1-D array
                self.Vn = ExtParam(model='Bus',
                                   src='Vn',
                                   indexer=self.Bus)

                self.Vn_mean = NumReduce(u=self.Vn,
                                         fun=np.mean,
                                         ref=self.Bus)

    Suppose we define two areas, 1 and 2, the Bus data looks like

        ===   =====  ====
        idx    area  Vn
        ---   -----  ----
        1      1     110
        2      2     220
        3      1     345
        4      1     500
        ===   =====  ====

    Then, `self.Bus.v` is a list of two lists ``[ [1, 3, 4], [2] ]``.
    `self.Vn.v` will be retrieved and linearly stored as ``[110, 345, 500,
    220]``. Based on the shape from `self.Bus`, :py:func:`numpy.mean` will be
    called on ``[110, 345, 500]`` and ``[220]`` respectively. Thus,
    `self.Vn_mean.v` will become ``[318.33, 220]``.

    """

    def __init__(self,
                 u,
                 ref: BackRef,
                 fun: Callable,
                 name=None,
                 tex_name=None,
                 info=None,
                 cache=True,
                 ):
        super().__init__(name=name, tex_name=tex_name, info=info)
        self.u = u
        self.ref = ref
        self.fun = fun
        self.cache = cache

    @property
    def v(self):
        """
        Return the reduced values from the reduction function in an array

        Returns
        -------
        The array ``self._v`` storing the reduced values
        """
        if self._v is not None and self.cache is True:
            return self._v

        if self._v is None:
            self._v = np.zeros(len(self.ref.v))

        idx = 0
        for i, v in enumerate(self.ref.v):
            self._v[i] = self.fun(self.u.v[idx:idx + len(v)])
            idx += len(v)
        return self._v


class NumRepeat(OperationService):
    r"""
    A helper Service type which repeats a v-provider's value based on the shape
    from a BackRef

    Examples
    --------
    NumRepeat was originally designed for computing the inertia-weighted average
    rotor speed (center of inertia speed). COI speed is computed with

    .. math ::
        \omega_{COI} = \frac{ \sum{M_i * \omega_i} } {\sum{M_i}}

    The numerator can be calculated with a mix of BackRef, ExtParam and
    ExtState. The denominator needs to be calculated with NumReduce and Service
    Repeat. That is, use NumReduce to calculate the sum, and use NumRepeat to
    repeat the summed value for each device.

    In the COI class, one would have

    .. code-block :: python

        class COIModel(...):
            def __init__(...):
                ...
                self.SynGen = BackRef()
                self.SynGenIdx = RefFlatten(ref=self.SynGen)
                self.M = ExtParam(model='SynGen',
                                  src='M',
                                  indexer=self.SynGenIdx)

                self.wgen = ExtState(model='SynGen',
                                     src='omega',
                                     indexer=self.SynGenIdx)

                self.Mt = NumReduce(u=self.M,
                                         fun=np.sum,
                                         ref=self.SynGen)

                self.Mtr = NumRepeat(u=self.Mt,
                                       ref=self.SynGen)

                self.pidx = IdxRepeat(u=self.idx,ref=self.SynGen)

    Finally, one would define the center of inertia speed as

    .. code-block :: python

        self.wcoi = Algeb(v_str='1', e_str='-wcoi')

        self.wcoi_sub = ExtAlgeb(model='COI',
                                 src='wcoi',
                                 e_str='M * wgen / Mtr',
                                 v_str='M / Mtr',
                                 indexer=self.pidx,
                                 )

    It is very worth noting that the implementation uses a trick to separate the
    average weighted sum into `n` sub-equations, each calculating the
    :math:`(M_i * \omega_i) / (\sum{M_i})`. Since all the variables are
    preserved in the sub-equation, the derivatives can be calculated correctly.

    """

    def __init__(self,
                 u,
                 ref,
                 **kwargs):
        super().__init__(**kwargs)
        self.u = u
        self.ref = ref

    @property
    def v(self):
        """
        Return the values of the repeated values in a sequential 1-D array

        Returns
        -------
        The array, ``self._v`` storing the repeated values
        """
        if self._v is None:
            self._v = np.zeros(len(list_flatten(self.ref.v)))

            idx = 0
            for i, v in enumerate(self.ref.v):
                self._v[idx:idx + len(v)] = self.u.v[i]
                idx += len(v)
            return self._v
        else:
            return self._v


class IdxRepeat(OperationService):
    """
    Helper class to repeat IdxParam.

    This class has the same functionality as
    :py:class:`andes.core.service.NumRepeat` but only operates on IdxParam,
    DataParam or NumParam.
    """

    def __init__(self,
                 u,
                 ref,
                 **kwargs):
        super().__init__(**kwargs)
        self.u = u
        self.ref = ref

    @property
    def v(self):
        if self._v is None:
            self._v = [''] * len(list_flatten(self.ref.v))
            idx = 0
            for i, v in enumerate(self.ref.v):
                for jj in range(idx, idx + len(v)):
                    self._v[jj] = self.u.v[i]
                idx += len(v)
            return self._v
        else:
            return self._v


class RefFlatten(OperationService):
    """
    A service type for flattening :py:class:`andes.core.service.BackRef` into a
    1-D list.

    Examples
    --------
    This class is used when one wants to pass `BackRef` values as indexer.

    :py:class:`andes.models.coi.COI` collects referencing
    :py:class:`andes.models.group.SynGen` with

    .. code-block :: python

        self.SynGen = BackRef(info='SynGen idx lists', export=False)

    After collecting BackRefs, `self.SynGen.v` will become a two-level list of
    indices, where the first level correspond to each COI and the second level
    correspond to generators of the COI.

    Convert `self.SynGen` into 1-d as `self.SynGenIdx`, which can be passed as
    indexer for retrieving other parameters and variables

    .. code-block :: python

        self.SynGenIdx = RefFlatten(ref=self.SynGen)

        self.M = ExtParam(model='SynGen', src='M',
                          indexer=self.SynGenIdx, export=False,
                          )
    """

    def __init__(self, ref, **kwargs):
        super().__init__(**kwargs)
        self.ref = ref

    @property
    def v(self):
        return list_flatten(self.ref.v)


class NumSelect(OperationService):
    """
    Class for selecting values for optional NumParam.

    NumSelect works with internal and external parameters.

    Examples
    --------
    One use case is to allow an optional turbine rating. One can do

    .. code:: python

        self.Tn = NumParam(default=None)
        self.Sg = ExtParam(...)
        self.Sn = DataSelect(Tn, Sg)

    """

    def __init__(self,
                 optional,
                 fallback,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 ):
        super().__init__(name=name, tex_name=tex_name, info=info)
        self.optional = optional
        self.fallback = fallback
        self._v = None

    @property
    def v(self):
        if self._v is None:
            self._v = [v1 if not np.isnan(v1)
                       else v2
                       for v1, v2 in zip(self.optional.v, self.fallback.v)]

            self._v = np.array(self._v)

        return self._v


class InitChecker(OperationService):
    """
    Class for checking init values against known typical values.

    Instances will be stored in `Model.services_post` and
    `Model.services_icheck`, which will be checked in `Model.post_init_check()`
    after initialization.

    Parameters
    ----------
    u
        v-provider to be checked
    lower : float, BaseParam, BaseVar, BaseService
        lower bound
    upper : float, BaseParam, BaseVar, BaseService
        upper bound
    equal : float, BaseParam, BaseVar, BaseService
        values that the value from `v_str` should equal
    not_equal : float, BaseParam, BaseVar, BaseService
        values that should not equal
    enable : bool
        True to enable checking

    Examples
    --------
    Let's say generator excitation voltages are known to be in the range of 1.6
    - 3.0 per unit. One can add the following instance to `GENBase` ::

        self._vfc = InitChecker(u=self.vf,
                                info='vf range',
                                lower=1.8,
                                upper=3.0,
                                )

    `lower` and `upper` can also take v-providers instead of float values.

    One can also pass float values from Config to make it adjustable as in our
    implementation of ``GENBase._vfc``.
    """

    def __init__(self, u, lower=None, upper=None, equal=None, not_equal=None,
                 enable=True, error_out=False, **kwargs):
        super().__init__(**kwargs)
        self.u = u
        self.lower = dummify(lower) if lower is not None else None
        self.upper = dummify(upper) if upper is not None else None
        self.equal = dummify(equal) if equal is not None else None
        self.not_equal = dummify(not_equal) if not_equal is not None else None
        self.enable = enable
        self.error_out = error_out

        self.checks = [(self.lower, np.less_equal, "out of typical lower limit", "limit"),
                       (self.upper, np.greater_equal, "out of typical upper limit", "limit"),
                       (self.equal, lambda a, b: np.logical_not(np.isclose(a, b)), 'should be equal', "expected"),
                       (self.not_equal, np.equal, 'should not be equal', "not expected")
                       ]

    def check(self):
        """
        Check the bounds and equality conditions.
        """
        if not self.enable:
            return

        if self._v is None:
            self._v = np.zeros_like(self.u.v)

        for check in self.checks:
            limit = check[0]
            func = check[1]
            text = check[2]
            text2 = check[3]
            if limit is None:
                continue

            self.v[:] = np.logical_or(self.v, func(self.u.v, limit.v))

            pos = np.argwhere(func(self.u.v, limit.v)).ravel()

            if len(pos) == 0:
                continue
            idx = [self.owner.idx.v[i] for i in pos]
            lim_v = limit.v * np.ones(self.n)

            title = f'{self.owner.class_name} {self.info} {text}.'

            err_dict = OrderedDict([('idx', idx),
                                    ('values', self.u.v[pos]),
                                    (f'{text2}', lim_v[pos]),
                                    ])
            data = list(map(list, zip(*err_dict.values())))

            tab = Tab(title=title, data=data, header=list(err_dict.keys()))
            if self.error_out:
                logger.error(tab.draw())
            else:
                logger.warning(tab.draw())

        self.v[:] = np.logical_not(self.v)


class FlagValue(BaseService):
    """
    Class for flagging values that equal to the given value.

    By default, values that equal to `value` will be flagged as `0`.
    Non-matching values will be flagged as `1`.

    Parameters
    ----------
    u
        Input parameter
    value
        Value to flag. Can be None, string, or a number.
    flag : 0 by default, only 0 or 1 is accepted.
        The flag for the matched ones

    Warnings
    --------
    `FlagNotNone` can only be applied to `BaseParam` with `cache=True`. Applying
    to `Service` will fail unless `cache` is False (at a performance cost).
    """

    def __init__(self, u, value, flag=0, name=None, tex_name=None, info=None, cache=True):
        BaseService.__init__(self, name=name, tex_name=tex_name, info=info)
        if flag != 0.0 and flag != 1.0:
            raise ValueError(f"flag must be 0 or 1. The given flag = {flag}.")

        self.u = u
        self.value = value
        self.flag = flag
        self.flag_neg = 1 - flag
        self.cache = cache

        self._v = None

    @property
    def v(self):
        new = False
        if self._v is None:
            self._v = np.zeros_like(self.u.v, dtype=float)
            new = True

        if not self.cache or new:
            # need to do it element-wise since `self.u.v` can be a list
            self._v[:] = np.array([self.flag if i == self.value else self.flag_neg
                                   for i in self.u.v])
        return self._v


class ApplyFunc(BaseService):
    """
    Class for applying a numerical function on a parameter..


    Warnings
    --------
    This class is not ready.

    Parameters
    ----------
    u
        Input parameter
    func
        A condition function that returns True or False.

    """

    def __init__(self, u, func, name=None, tex_name=None, info=None, cache=True):
        BaseService.__init__(self, name=name, tex_name=tex_name, info=info)
        self.u = u
        self.func = func
        self.cache = cache
        self._v = None
        self._eval = False  # has been evaluated previously

    @property
    def v(self):
        if not self._eval:
            self._v = np.zeros_like(self.u.v, dtype=float)

        if not self.cache or (not self._eval):
            self._v[:] = self.func(self.u.v)

        self._eval = True
        return self._v


class FlagCondition(BaseService):
    """
    Class for flagging values based on a condition function.

    By default, values whose condition function output equal that equal to
    True/1 will be flagged as `1`. `0` otherwise.

    Parameters
    ----------
    u
        Input parameter
    func
        A condition function that returns True or False.
    flag : 1 by default, only 0 or 1 is accepted.
        The flag for the inputs whose condition output is True.

    Warnings
    --------
    This class is not ready.

    `FlagCondition` can only be applied to `BaseParam` with `cache=True`.
    Applying to `Service` will fail unless `cache` is False (at a performance
    cost).
    """

    def __init__(self, u, func, flag=1, name=None, tex_name=None, info=None, cache=True):
        BaseService.__init__(self, name=name, tex_name=tex_name, info=info)
        if flag != 0.0 and flag != 1.0:
            raise ValueError(f"flag must be 0 or 1. The given flag = {flag}.")

        self.u = u
        self.func = func
        self.flag = flag
        self.flag_neg = 1 - flag
        self.cache = cache

        self._v = None
        self._eval = False  # has been evaluated previously

    @property
    def v(self):
        if not self._eval:
            self._v = np.zeros_like(self.u.v, dtype=float)

        if not self.cache or (not self._eval):
            cond_out = self.func(self.u.v)

            self._v[:] = np.array([self.flag if i == 1 else self.flag_neg
                                   for i in cond_out])

        self._eval = True
        return self._v


class FlagLessThan(FlagCondition):
    """
    Service for flagging parameters < or <= the given value element-wise.

    Parameters that satisfy the comparison (u < or <= value) will flagged as
    `flag` (1 by default).
    """

    def __init__(self, u, value=0.0, flag=1, equal=False,
                 name=None, tex_name=None, info=None, cache=True):

        self.value = dummify(value)
        self.equal = equal

        if self.equal is True:
            self.func = lambda x: np.less_equal(x, self.value.v)
        else:
            self.func = lambda x: np.less(x, self.value.v)

        FlagCondition.__init__(self, u, func=self.func,
                               flag=flag, name=name,
                               tex_name=tex_name, info=info, cache=cache,
                               )


class FlagGreaterThan(FlagCondition):
    """
    Service for flagging parameters > or >= the given value element-wise.

    Parameters that satisfy the comparison (u > or >= value) will flagged as
    `flag` (1 by default).
    """

    def __init__(self, u, value=0.0, flag=1, equal=False,
                 name=None, tex_name=None, info=None, cache=True):

        self.value = dummify(value)
        self.equal = equal

        if self.equal is True:
            self.func = lambda x: np.greater_equal(x, self.value.v)
        else:
            self.func = lambda x: np.greater(x, self.value.v)

        FlagCondition.__init__(self, u, func=self.func,
                               flag=flag, name=name,
                               tex_name=tex_name, info=info, cache=cache,
                               )


class CurrentSign(ConstService):
    """
    Service for computing the sign of the current flowing through a series
    device.

    With a given line connecting `bus1` and `bus2`, one can compute the current
    flow using ``(v1*exp(1j*a1) - v2*exp(1j*a2)) / (r + jx)`` whose value is the
    outflow on `bus1`.

    `CurrentSign` can be used to compute the sign to be multiplied depending on
    the observing bus. For each value in `bus`, the sign will be ``+1`` if it
    appears in `bus1` or ``-1`` otherwise.

    ::

        bus1          bus2
         *------>>-----*
        bus(+)        bus(-)

    """

    def __init__(self, bus, bus1, bus2,  name=None, tex_name=None, info=None):
        ConstService.__init__(self, v_numeric=self.check, name=name, tex_name=tex_name, info=info)
        self.bus = bus
        self.bus1 = bus1
        self.bus2 = bus2

    def check(self, **kwargs):
        out = np.zeros_like(self.v)

        for idx, (bus, bus1, bus2) in enumerate(zip(self.bus.v, self.bus1.v, self.bus2.v)):
            if bus == bus1:
                out[idx] = 1
            elif bus == bus2:
                out[idx] = -1
            else:
                raise ValueError(f"bus {bus} is not terminal of the line connecting {bus1} and {bus2}. "
                                 f"Check the data of {self.bus.owner.class_name}.{self.bus.name}")

        return out


class Replace(BaseService):
    """
    Replace parameters with new values if the function returns True
    """

    def __init__(self, old_val, flt, new_val, name=None, tex_name=None, info=None, cache=True):
        BaseService.__init__(self, name=name, tex_name=tex_name, info=info)
        self.cache = cache
        self.filter = flt  # function
        self.old_val = old_val
        self.new_val = dummify(new_val)
        self._v = None

    @property
    def v(self):
        new = False
        if self._v is None or not self.cache:
            self._v = np.zeros_like(self.old_val.v, dtype=float)
            new = True

        if not self.cache or new:
            new_v = self.new_val.v * np.ones_like(self.old_val.v)
            flt = self.filter(self.old_val.v)
            self._v[:] = new_v * flt + self.old_val.v * (1 - flt)

        return self._v


class ParamCalc(BaseService):
    """
    Parameter calculation service.

    Useful to create parameters calculated instantly from existing ones.
    """

    def __init__(self, param1, param2, func, name=None, tex_name=None, info=None,
                 cache=True):
        BaseService.__init__(self, name=name, tex_name=tex_name, info=info)
        self.param1 = param1
        self.param2 = param2
        self.func = func
        self.cache = cache
        self._v = None

    @property
    def v(self):
        new = False
        if self._v is None:
            new = True
            self._v = np.zeros_like(self.param1.v, dtype=float)

        if not self.cache or new:
            self._v[:] = self.func(self.param1.v,
                                   self.param2.v)

        return self._v


class RandomService(BaseService):
    """
    A service type for generating random numbers.

    Parameters
    ----------
    name : str
        Name
    func : Callable
        A callable for generating the random variable.

    Warnings
    --------
    The value will be randomized every time it is accessed. Do not use it if the
    value needs to be stable for each simulation step.
    """

    def __init__(self, func=np.random.rand, **kwargs):
        super(RandomService, self).__init__(**kwargs)
        self.func = func

    @property
    def v(self):
        """
        This class has `v` wrapped by a property decorator.

        Returns
        -------
        array-like
            Randomly generated service variables
        """
        return np.random.rand(self.n)


class SwBlock(OperationService):
    """
    Service type for switched shunt blocks.
    """

    def __init__(self, *, init, ns, blocks, ext_sel=None,
                 name=None, tex_name=None, info=None):
        OperationService.__init__(self, name=name, tex_name=tex_name, info=info)
        self.init = init
        self.ns = ns
        self.bs = blocks

        self.sel = None
        self.bcs = None
        self.maxsel = None
        self.ext_sel = ext_sel

    @property
    def v(self):
        self.check_data()

        # allocate memory
        if self._v is not None:
            return self._v

        # initialize
        n_dev = len(self.init.v)
        self._v = np.array(self.init.v)           # effective value
        self.sel = np.zeros(n_dev, dtype=int)     # the index of capacity in use
        self.bcs = [0] * n_dev                    # cumulative sums of `bs`
        self.maxsel = np.zeros(n_dev, dtype=int)  # max index into `bs`

        for idx in range(n_dev):
            item = self.ns.v[idx]
            self.maxsel[idx] = sum(item)

        # repeat each value in `bs` by `ns` times
        for idx in range(n_dev):
            # disabled - use default `b`
            if self.maxsel[idx] == 0:
                self.bcs[idx] = np.array([self.init.v[idx]])
                continue

            # calculate cumulative sum
            b_rep = list([0])
            for nn, bb in zip(self.ns.v[idx], self.bs.v[idx]):
                b_rep += [bb] * nn
            self.bcs[idx] = np.cumsum(b_rep)

        if self.ext_sel is None:
            # use internal selector - for shunt's b attribute
            self.find_sel()
        else:
            # use external selector - for shunt's g attribute
            self.sel = self.ext_sel.sel  # modify reference

        self.set_v()
        return self._v

    def check_data(self):
        """
        Check data consistency.
        """
        model = self.ns.owner.class_name
        for idx in range(len(self.ns.v)):
            device = self.ns.owner.idx.v[idx]
            bs_name = self.bs.name
            ns_name = self.ns.name

            if isinstance(self.ns.v[idx], (list, np.ndarray)):
                if len(self.ns.v[idx]) == 0:
                    continue
                if len(self.ns.v[idx]) == 1 and self.ns.v[idx][0] == 0:
                    continue

            if isinstance(self.bs.v[idx], (int, float, str)):
                raise ValueError("<%s>: idx=%s, `%s` parameter should be list literal, got %s" %
                                 (model, device, bs_name, self.bs.v[idx]))

            if len(self.ns.v[idx]) != len(self.bs.v[idx]):
                raise ValueError("<%s>: idx=%s, `%s` and `%s` lengths do not match" %
                                 (model, device, bs_name, ns_name))

    def adjust(self, amount):
        """
        Adjust capacitor banks by an amount.
        """
        if self.ext_sel is None:
            self.sel[:] += amount

        self.set_v()

    def set_v(self):
        """
        Set values to `_v` based on `sel`.
        """

        for idx in range(len(self._v)):
            self._v[idx] = self.bcs[idx][self.sel[idx]]

    def find_sel(self):
        """
        Determine the initial shunt selection level.
        """
        for idx in range(len(self._v)):
            binit = self.init.v[idx]
            if binit > self.bcs[idx][-1]:
                # out of maximum b
                logger.warning("<%s> idx=%s, initial %s=%g is greater than max=%g",
                               self.init.owner.class_name,
                               self.init.owner.idx.v[idx],
                               self.init.name, binit,
                               self.bcs[idx][-1])
                self.sel[idx] = self.maxsel[idx]

            elif binit == self.bcs[idx][-1]:
                self.sel[idx] = self.maxsel[idx]

            elif binit < 0:
                logger.warning("<%s> idx=%s, initial %s=%g is less than zero",
                               self.init.owner.class_name,
                               self.init.owner.idx.v[idx],
                               self.init.name, binit)
                self.sel[idx] = 0

            else:
                for pos in range(self.maxsel[idx]):
                    blo = self.bcs[idx][pos]
                    bup = self.bcs[idx][pos + 1]
                    if binit == blo:
                        self.sel[idx] = pos
                        break
                    if binit == bup:
                        self.sel[idx] = pos + 1
                        break
                    if blo < binit < bup:
                        self.sel[idx] = pos + 1
                        break



"""
R, L, C and their combinations.
"""

from andes.core import NumParam, Algeb, State
from andes.models.dc.dcbase import DC2Term


class R(DC2Term):
    """
    Resistive dc line
    """

    def __init__(self, system, config):
        DC2Term.__init__(self, system, config)
        self.flags.pflow = True
        self.group = 'DCLink'
        self.R = NumParam(unit='p.u.',
                          info='DC line resistance',
                          non_zero=True,
                          default=0.01,
                          r=True,
                          )
        self.Idc = Algeb(tex_name='I_{dc}',
                         info='Current from node 2 to 1',
                         unit='p.u.',
                         v_str='u * (v2 - v1) / R',
                         e_str='u * (v2 - v1) / R - Idc',
                         )
        self.v1.e_str = '-Idc'
        self.v2.e_str = '+Idc'


class L(DC2Term):
    """
    Inductive dc line
    """

    def __init__(self, system, config):
        DC2Term.__init__(self, system, config)
        self.flags.pflow = True
        self.group = 'DCLink'

        self.L = NumParam(unit='p.u.',
                          info='DC line inductance',
                          non_zero=True,
                          default=0.001,
                          r=True,
                          )
        self.IL = State(tex_name='I_L',
                        info='Inductance current',
                        unit='p.u.',
                        v_str='0',
                        e_str='-u * (v1 - v2)',
                        t_const=self.L,
                        )
        self.v1.e_str = '-IL'
        self.v2.e_str = '+IL'


class C(DC2Term):
    """
    Capacitive dc branch
    """

    def __init__(self, system, config):
        DC2Term.__init__(self, system, config)
        self.flags.pflow = True
        self.group = 'DCLink'

        self.C = NumParam(unit='p.u.',
                          info='DC capacitance',
                          non_zero=True,
                          default=0.001,
                          g=True,
                          )
        self.vC = State(tex_name='v_C',
                        info='Capacitor current',
                        unit='p.u.',
                        v_str='0',
                        e_str='-u * Idc',
                        t_const=self.C
                        )
        self.Idc = Algeb(tex_name='I_{dc}',
                         info='Current from node 2 to 1',
                         unit='p.u.',
                         v_str='0',
                         e_str='u * (vC - (v1 - v2)) + '
                               '(1 - u) * Idc',
                         diag_eps=True,
                         )
        self.v1.e_str = '-Idc'
        self.v2.e_str = '+Idc'


class RLs(DC2Term):
    def __init__(self, system, config):
        DC2Term.__init__(self, system, config)
        self.flags.pflow = True
        self.group = 'DCLink'

        self.R = NumParam(unit='p.u.',
                          tex_name='R',
                          info='DC line resistance',
                          non_zero=True,
                          default=0.01,
                          r=True,
                          )
        self.L = NumParam(unit='p.u.',
                          tex_name='L',
                          info='DC line inductance',
                          non_zero=True,
                          default=0.001,
                          r=True,
                          )
        self.IL = State(tex_name='I_L',
                        info='Inductance current',
                        unit='p.u.',
                        e_str='u * (v1 - v2 - R * IL)',
                        v_str='(v1 - v2) / R',
                        t_const=self.L,
                        )
        self.Idc = Algeb(tex_name='I_{dc}',
                         info='Current from node 2 to 1',
                         unit='p.u.',
                         e_str='-u * IL - Idc',
                         v_str='-u * (v1 - v2) / R',
                         )
        self.v1.e_str = '-Idc'
        self.v2.e_str = '+Idc'


class RCp(DC2Term):
    def __init__(self, system, config):
        DC2Term.__init__(self, system, config)
        self.flags.pflow = True
        self.group = 'DCLink'

        self.R = NumParam(unit='p.u.',
                          tex_name='R',
                          info='DC line resistance',
                          non_zero=True,
                          default=0.01,
                          r=True,
                          )
        self.C = NumParam(unit='p.u.',
                          tex_name='C',
                          info='DC capacitance',
                          non_zero=True,
                          default=0.001,
                          g=True,
                          )
        self.vC = State(tex_name='v_C',
                        info='Capacitor current',
                        unit='p.u.',
                        e_str='-u * (Idc - vC/R)',
                        v_str='v1 - v2',
                        t_const=self.C,
                        )
        self.Idc = Algeb(tex_name='I_{dc}',
                         info='Current from node 2 to 1',
                         unit='p.u.',
                         e_str='u * (vC - (v1 - v2)) + '
                               '(1 - u) * Idc',
                         v_str='-(v1 - v2) / R',
                         diag_eps=True,
                         )
        self.v1.e_str = '-Idc'
        self.v2.e_str = '+Idc'


class RLCp(DC2Term):
    def __init__(self, system, config):
        DC2Term.__init__(self, system, config)
        self.flags.pflow = True
        self.group = 'DCLink'

        self.R = NumParam(unit='p.u.',
                          tex_name='R',
                          info='DC line resistance',
                          non_zero=True,
                          default=0.01,
                          r=True,
                          )
        self.L = NumParam(unit='p.u.',
                          tex_name='L',
                          info='DC line inductance',
                          non_zero=True,
                          default=0.001,
                          r=True,
                          )
        self.C = NumParam(unit='p.u.',
                          tex_name='C',
                          info='DC capacitance',
                          non_zero=True,
                          default=0.001,
                          g=True,
                          )
        self.IL = State(tex_name='I_L',
                        info='Inductance current',
                        unit='p.u.',
                        v_str='0',
                        e_str='u * vC',
                        t_const=self.L,
                        )
        self.vC = State(tex_name='v_C',
                        info='Capacitor current',
                        unit='p.u.',
                        e_str='-u * (Idc - vC/R - IL)',
                        v_str='v1 - v2',
                        t_const=self.C,
                        )
        self.Idc = Algeb(tex_name='I_{dc}',
                         info='Current from node 2 to 1',
                         unit='p.u.',
                         e_str='u * (vC - (v1 - v2)) + '
                               '(1 - u) * Idc',
                         v_str='-(v1 - v2) / R',
                         diag_eps=True,
                         )
        self.v1.e_str = '-Idc'
        self.v2.e_str = '+Idc'


class RCs(DC2Term):
    def __init__(self, system, config):
        DC2Term.__init__(self, system, config)
        self.flags.pflow = True
        self.group = 'DCLink'

        self.R = NumParam(unit='p.u.',
                          tex_name='R',
                          info='DC line resistance',
                          non_zero=True,
                          default=0.01,
                          r=True,
                          )
        self.C = NumParam(unit='p.u.',
                          tex_name='C',
                          info='DC capacitance',
                          non_zero=True,
                          default=0.001,
                          g=True,
                          )
        self.vC = State(tex_name='v_C',
                        info='Capacitor current',
                        unit='p.u.',
                        e_str='-u * Idc',
                        v_str='v1 - v2',
                        t_const=self.C,
                        )
        self.Idc = Algeb(tex_name='I_{dc}',
                         info='Current from node 2 to 1',
                         unit='p.u.',
                         e_str='u * (vC - (v1 - v2) - Idc * R) + '
                               '(1 - u) * Idc',
                         v_str='-(v1 - v2) / R',
                         diag_eps=True,
                         )
        self.v1.e_str = '-Idc'
        self.v2.e_str = '+Idc'


class RLCs(DC2Term):
    def __init__(self, system, config):
        DC2Term.__init__(self, system, config)
        self.flags.pflow = True
        self.group = 'DCLink'

        self.R = NumParam(unit='p.u.',
                          tex_name='R',
                          info='DC line resistance',
                          non_zero=True,
                          default=0.01,
                          r=True,
                          )
        self.L = NumParam(unit='p.u.',
                          tex_name='L',
                          info='DC line inductance',
                          non_zero=True,
                          default=0.001,
                          r=True,
                          )
        self.C = NumParam(unit='p.u.',
                          tex_name='C',
                          info='DC capacitance',
                          non_zero=True,
                          default=0.001,
                          g=True,
                          )
        self.IL = State(tex_name='I_L',
                        info='Inductance current',
                        unit='p.u.',
                        e_str='u * (v1 - v2 - R * IL - vC)',
                        v_str='0',
                        t_const=self.L,
                        )
        self.vC = State(tex_name='v_C',
                        info='Capacitor current',
                        unit='p.u.',
                        e_str='u * IL',
                        v_str='v1 - v2',
                        t_const=self.C,
                        )
        self.Idc = Algeb(tex_name='I_{dc}',
                         info='Current from node 2 to 1',
                         unit='p.u.',
                         e_str='-IL - Idc',
                         v_str='0',
                         diag_eps=True,
                         )
        self.v1.e_str = '-Idc'
        self.v2.e_str = '+Idc'



from andes.core.block import Block
from andes.core.common import dummify
from andes.core.service import ConstService, FlagValue, InitChecker


class ExcExpSat(Block):
    r"""
    Exponential exciter saturation block to calculate
    A and B from E1, SE1, E2 and SE2.
    Input parameters will be corrected and the user will be warned.
    To disable saturation, set either E1 or E2 to 0.

    Parameters
    ----------
    E1 : BaseParam
        First point of excitation field voltage
    SE1: BaseParam
        Coefficient corresponding to E1
    E2 : BaseParam
        Second point of excitation field voltage
    SE2: BaseParam
        Coefficient corresponding to E2
    """

    def __init__(self, E1, SE1, E2, SE2, name=None, tex_name=None, info=None):
        Block.__init__(self, name=name, tex_name=tex_name, info=info)

        self._E1 = E1
        self._E2 = E2
        self._SE1 = SE1
        self._SE2 = SE2

        self.zE1 = FlagValue(self._E1, value=0.,
                             info='Flag non-zeros in E1',
                             tex_name='z^{E1}',
                             )
        self.zE2 = FlagValue(self._E2, value=0.,
                             info='Flag non-zeros in E2',
                             tex_name='z^{E2}',
                             )
        self.zSE1 = FlagValue(self._SE1, value=0.,
                              info='Flag non-zeros in SE1',
                              tex_name='z^{SE1}',
                              )
        self.zSE2 = FlagValue(self._SE2, value=0.,
                              info='Flag non-zeros in SE2',
                              tex_name='z^{SE2}')

        # disallow E1 = E2 != 0 since the curve fitting will fail
        self.E12c = InitChecker(
            self._E1, not_equal=self._E2,
            info='E1 and E2 after correction',
            error_out=True,
        )

        # data correction for E1, E2, SE1
        self.E1 = ConstService(
            tex_name='E^{1c}',
            info='Corrected E1 data',
        )
        self.E2 = ConstService(
            tex_name='E^{2c}',
            info='Corrected E2 data',
        )
        self.SE1 = ConstService(
            tex_name='SE^{1c}',
            info='Corrected SE1 data',
        )
        self.SE2 = ConstService(
            tex_name='SE^{2c}',
            info='Corrected SE2 data',
        )
        self.A = ConstService(info='Saturation gain',
                              tex_name='A^e',
                              )
        self.B = ConstService(info='Exponential coef. in saturation',
                              tex_name='B^e',
                              )
        self.vars = {
            'E1': self.E1,
            'E2': self.E2,
            'SE1': self.SE1,
            'SE2': self.SE2,
            'zE1': self.zE1,
            'zE2': self.zE2,
            'zSE1': self.zSE1,
            'zSE2': self.zSE2,
            'A': self.A,
            'B': self.B,
        }

    def define(self):
        r"""
        Notes
        -----
        The implementation solves for coefficients `A` and `B`
        which satisfy

        .. math ::
            E_1  S_{E1} = A e^{E1\times B}
            E_2  S_{E2} = A e^{E2\times B}

        The solutions are given by

        .. math ::
            E_{1} S_{E1} e^{ \frac{E_1 \log{ \left( \frac{E_2 S_{E2}} {E_1 S_{E1}} \right)} } {E_1 - E_2}}
            - \frac{\log{\left(\frac{E_2 S_{E2}}{E_1 S_{E1}} \right)}}{E_1 - E_2}
        """
        self.E1.v_str = f'{self._E1.name} + (1 - {self.name}_zE1)'
        self.E2.v_str = f'{self._E2.name} + 2*(1 - {self.name}_zE2)'

        self.SE1.v_str = f'{self._SE1.name} + (1 - {self.name}_zSE1)'
        self.SE2.v_str = f'{self._SE2.name} + 2*(1 - {self.name}_zSE2)'

        self.A.v_str = f'{self.name}_zE1*{self.name}_zE2 * ' \
                       f'{self.name}_E1*{self.name}_SE1*' \
                       f'exp({self.name}_E1*log({self.name}_E2*{self.name}_SE2/' \
                       f'({self.name}_E1*{self.name}_SE1))/({self.name}_E1-{self.name}_E2))'

        self.B.v_str = f'-log({self.name}_E2*{self.name}_SE2/({self.name}_E1*{self.name}_SE1))/' \
                       f'({self.name}_E1 - {self.name}_E2)'


class ExcQuadSat(Block):
    r"""
    Exponential exciter saturation block to calculate
    A and B from E1, SE1, E2 and SE2.
    Input parameters will be corrected and the user will be warned.
    To disable saturation, set either E1 or E2 to 0.

    Parameters
    ----------
    E1 : BaseParam
        First point of excitation field voltage
    SE1: BaseParam
        Coefficient corresponding to E1
    E2 : BaseParam
        Second point of excitation field voltage
    SE2: BaseParam
        Coefficient corresponding to E2
    """

    def __init__(self, E1, SE1, E2, SE2, name=None, tex_name=None, info=None):
        Block.__init__(self, name=name, tex_name=tex_name, info=info)

        self._E1 = dummify(E1)
        self._E2 = dummify(E2)
        self._SE1 = SE1
        self._SE2 = SE2

        self.zSE2 = FlagValue(self._SE2, value=0.,
                              info='Flag non-zeros in SE2',
                              tex_name='z^{SE2}')

        # data correction for E1, E2, SE1 (TODO)
        self.E1 = ConstService(
            tex_name='E^{1c}',
            info='Corrected E1 data',
        )
        self.E2 = ConstService(
            tex_name='E^{2c}',
            info='Corrected E2 data',
        )
        self.SE1 = ConstService(
            tex_name='SE^{1c}',
            info='Corrected SE1 data',
        )
        self.SE2 = ConstService(
            tex_name='SE^{2c}',
            info='Corrected SE2 data',
        )
        self.a = ConstService(info='Intermediate Sat coeff',
                              tex_name='a',
                              )
        self.A = ConstService(info='Saturation start',
                              tex_name='A^q',
                              )
        self.B = ConstService(info='Saturation gain',
                              tex_name='B^q',
                              )
        self.vars = {
            'E1': self.E1,
            'E2': self.E2,
            'SE1': self.SE1,
            'SE2': self.SE2,
            'zSE2': self.zSE2,
            'a': self.a,
            'A': self.A,
            'B': self.B,
        }

    def define(self):
        r"""
        Notes
        -----
        TODO.
        """

        self.E1.v_str = f'{self._E1.name}'
        self.E2.v_str = f'{self._E2.name}'
        self.SE1.v_str = f'{self._SE1.name}'
        self.SE2.v_str = f'{self._SE2.name} + 2 * (1 - {self.name}_zSE2)'

        self.a.v_str = f'(Indicator({self.name}_SE2>0)+Indicator({self.name}_SE2<0)) * ' \
                       f'sqrt({self.name}_SE1 * {self.name}_E1 /({self.name}_SE2 * {self.name}_E2))'

        self.A.v_str = f'{self.name}_E2 - ({self.name}_E1 - {self.name}_E2) / ({self.name}_a - 1)'

        self.B.v_str = f'(Indicator({self.name}_a>0)+Indicator({self.name}_a<0)) *' \
                       f'{self.name}_SE2 * {self.name}_E2 * ({self.name}_a - 1)**2 / ' \
                       f'({self.name}_E1 - {self.name}_E2)** 2'



from andes.core import Lag, LeadLag, NumParam
from andes.core.block import GainLimiter
from andes.models.governor.tgbase import TGBase, TGBaseData


class IEESGOData(TGBaseData):
    def __init__(self):
        TGBaseData.__init__(self)
        self.T1 = NumParam(info='Controller lag',
                           default=0.02,
                           tex_name='T_1',
                           vrange=(0, 100),
                           )
        self.T2 = NumParam(info='Lead compensation',
                           default=1.0,
                           tex_name='T_2',
                           vrange=(0, 10),
                           )
        self.T3 = NumParam(info='Governor lag',
                           default=1.0,
                           tex_name='T_3',
                           vrange=(0.04, 1.0),
                           )
        self.T4 = NumParam(info='Steam inlet delay',
                           default=0.5,
                           tex_name='T_4',
                           vrange=(0.0, 1.0),
                           )
        self.T5 = NumParam(info='Reheater delay',
                           default=10.0,
                           tex_name='T_5',
                           vrange=(0.0, 50.0),
                           )
        self.T6 = NumParam(info='Crossover delay',
                           default=0.5,
                           tex_name='T_6',
                           vrange=(0.0, 1.0),
                           )

        self.K1 = NumParam(info='1/pu regulation',
                           default=0.02,
                           tex_name='K_1',
                           vrange=(5, 30),
                           )
        self.K2 = NumParam(info='fraction K2',
                           default=1.0,
                           tex_name='K_2',
                           vrange=(0, 3),
                           )
        self.K3 = NumParam(info='fraction K3',
                           default=1.0,
                           tex_name='K_3',
                           vrange=(-1.0, 1.0),
                           )

        self.PMAX = NumParam(default=5, tex_name='P_{MAX}',
                             info='Max. turbine power',
                             vrange=(0.5, 1.5), power=True,
                             )
        self.PMIN = NumParam(default=0., tex_name='P_{MIN}',
                             info='Min. turbine power',
                             vrange=(0.0, 0.5), power=True,
                             )


class IEESGOModel(TGBase):
    def __init__(self, system, config):
        TGBase.__init__(self, system, config)

        self.F1 = Lag(u='ue * (omega - wref)',
                      T=self.T1,
                      K=self.K1,
                      )

        self.F2 = LeadLag(u=self.F1_y,
                          T1=self.T2,
                          T2=self.T3,
                          K=1.0,
                          )

        self.HL = GainLimiter(u='ue * (paux + pref0 - F2_y)',
                              K=1.0,
                              R=1.0,
                              lower=self.PMIN,
                              upper=self.PMAX,
                              )
        self.F3 = Lag(u=self.HL_y, T=self.T4, K=1.0,
                      )

        self.F4 = Lag(u=self.F3_y, T=self.T5, K=self.K2,
                      )

        self.F5 = Lag(u=self.F4_y, T=self.T6, K=self.K3,
                      )

        self.pout.e_str = 'ue * ((1-K2)*F3_y + (1-K3)*F4_y + F5_y) - pout'


class IEESGO(IEESGOData, IEESGOModel):
    """
    IEEE Standard Governor (IEESGO).
    """

    def __init__(self, system, config):
        IEESGOData.__init__(self)
        IEESGOModel.__init__(self, system, config)



from andes.core import ModelData, Model, IdxParam, NumParam, ExtAlgeb, Algeb


class Ground(ModelData, Model):
    """
    Ground model that sets the voltage of the connected DC node.
    """

    def __init__(self, system, config):
        ModelData.__init__(self)
        self.node = IdxParam(default=None,
                             info='Node index',
                             mandatory=True,
                             model='Node',
                             )
        self.voltage = NumParam(default=0.0,
                                tex_name='V_0',
                                info='Ground voltage (typically 0)',
                                unit='p.u.',
                                )
        Model.__init__(self, system, config)
        self.flags.update({'pflow': True})
        self.group = 'DCLink'
        self.v = ExtAlgeb(model='Node',
                          src='v',
                          indexer=self.node,
                          e_str='-Idc',
                          ename='-Idc',
                          tex_ename='-I_{dc}',
                          )
        self.Idc = Algeb(tex_name='I_{dc}',
                         info='Ficticious current injection from ground',
                         e_str='u * (v - voltage)',
                         v_str='0',
                         diag_eps=True,
                         )
        self.v.e_str = '-Idc'



from collections import OrderedDict

from andes.core.discrete import Sampling
from andes.core.model import Model, ModelData
from andes.core.param import ExtParam, IdxParam, NumParam
from andes.core.service import BackRef, ConstService, DeviceFinder
from andes.core.var import Algeb, ExtAlgeb
from andes.utils.tab import Tab


class AreaData(ModelData):
    def __init__(self):
        super().__init__()


class Area(AreaData, Model):
    """
    Area model.

    Area collects back references from the Bus model and
    the ACTopology group.
    """
    def __init__(self, system, config):
        AreaData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'Collection'
        self.flags.pflow = True
        self.flags.tds = True

        self.Bus = BackRef()
        self.ACTopology = BackRef()

    def bus_table(self):
        """
        Return a formatted table with area idx and bus idx correspondence

        Returns
        -------
        str
            Formatted table

        """
        if self.n:
            header = ['Area ID', 'Bus ID']
            rows = [(i, j) for i, j in zip(self.idx.v, self.Bus.v)]
            return Tab(header=header, data=rows).draw()
        else:
            return ''


class ACEData(ModelData):
    """
    Area Control Error data
    """

    def __init__(self):
        ModelData.__init__(self)
        self.bus = IdxParam(model='Bus', info="bus idx for freq. measurement", mandatory=True)
        self.bias = NumParam(default=1.0, info='bias parameter', tex_name=r'\beta',
                             unit='MW/0.1Hz', power=True)

        self.busf = IdxParam(info='Optional BusFreq device idx', model='BusFreq',
                             default=None)


class ACEc(ACEData, Model):
    """
    Area Control Error model.

    Continuous frequency sampling.
    System base frequency from ``system.config.freq`` is used.

    Note: area idx is automatically retrieved from `bus`.
    """

    def __init__(self, system, config):
        ACEData.__init__(self)
        Model.__init__(self, system, config)

        self.flags.tds = True
        self.group = 'Calculation'

        self.config.add(OrderedDict([('freq_model', 'BusFreq'),
                                     ]))
        self.config.add_extra('_help',
                              {'freq_model': 'default freq. measurement model',
                               })
        self.config.add_extra('_alt', {'freq_model': ('BusFreq',)})

        self.area = ExtParam(model='Bus', src='area', indexer=self.bus, export=False)

        self.busf.model = self.config.freq_model
        self.busfreq = DeviceFinder(self.busf, link=self.bus, idx_name='bus',
                                    default_model='BusFreq')

        self.imva = ConstService(v_str='1/sys_mva', info='reciprocal of system mva',
                                 tex_name='1/S_{b, sys}')

        self.f = ExtAlgeb(model='FreqMeasurement',
                          src='f',
                          indexer=self.busfreq,
                          export=False,
                          info='Bus frequency',
                          unit='p.u. (Hz)'
                          )
        self.ace = Algeb(info='area control error',
                         unit='p.u. (MW)',
                         tex_name='ace',
                         e_str='10 * (bias * imva) * sys_f * (f - 1) - ace',
                         )


class ACE(ACEc):
    """
    Area Control Error model.

    Discrete frequency sampling.
    System base frequency from ``system.config.freq`` is used.

    Frequency sampling period (in seconds) can be specified in
    ``ACE.config.interval``. The sampling start time (in seconds)
    can be specified in ``ACE.config.offset``.

    Note: area idx is automatically retrieved from `bus`.
    """

    def __init__(self, system, config):
        ACEc.__init__(self, system, config)

        self.config.add(OrderedDict([('interval', 4.0),
                                     ('offset', 0.0),
                                     ]))
        self.config.add_extra('_help', {'interval': 'sampling time interval',
                                        'offset': 'sampling time offset'})

        self.fs = Sampling(self.f,
                           interval=self.config.interval,
                           offset=self.config.offset,
                           tex_name='f_s',
                           info='Sampled freq.',
                           )

        self.ace.e_str = '10 * (bias * imva) * sys_f * (fs_v - 1) - ace'



"""
PSSE RAW parser from em_psse

https://github.com/anderson-optimization/em-psse

License Pending
"""
import os
from io import StringIO
from andes.shared import pd, yaml

import logging

logger = logging.getLogger(__name__)

dirname = os.path.dirname(__file__)
with open('{}/psse-modes.yaml'.format(dirname), 'r') as in_file:
    modes = yaml.full_load(in_file)


def get_signals(line_num, line, current_mode):
    # print(line_num,line)
    signals = []
    for m in modes:
        for s in m['signal']:
            if 'text' in s:
                if s['text'] in line:
                    signals.append((s, m))
            if 'line' in s:
                if s['line'] == line_num:
                    signals.append((s, m))
    return signals


def read_transformer(lines, records):
    iter_lines = iter(lines)
    # records=[]
    headers = [r['header'] for r in records]
    line_holder = {
        2: [
            [headers[0]],
            [headers[1]],
            [headers[2]],
            [headers[3]]
        ],
        3: [
            [headers[0]],
            [headers[1]],
            [headers[2]],
            [headers[3]],
            [headers[4]]
        ]
    }
    for line in iter_lines:
        rows = [line]
        rows.append(next(iter_lines))
        rows.append(next(iter_lines))
        rows.append(next(iter_lines))
        windings = None
        try:
            flag = int(line.split(',')[2])
            if flag != 0:
                raise ValueError('Not a 2 phase winding')
            windings = 2
        except Exception:
            windings = 3
            rows.append(next(iter_lines))

        for r in range(len(rows)):
            line_holder[windings][r].append(rows[r])

    dfs = {
        2: [],
        3: []
    }
    for winding in line_holder:
        rc = 0
        for record in line_holder[winding]:
            rc += 1
            text = StringIO(''.join(record))
            dfs[winding].append(pd.read_table(text, sep=',', error_bad_lines=False))
            logger.debug("{} {} {}".format(winding, rc, len(record)))
            logger.debug("{}".format(record[0]))
    return dfs


def read_twodc(lines, records):
    iter_lines = iter(lines)
    # records=[]
    headers = [r['header'] for r in records]
    line_holder = [
        [headers[0]],
        [headers[1]],
        [headers[2]]
    ]
    for line in iter_lines:
        rows = [line]
        rows.append(next(iter_lines))
        rows.append(next(iter_lines))

        for r in range(len(rows)):
            line_holder[r].append(rows[r])

    dfs = []
    rc = 0
    for record in line_holder:
        rc += 1
        text = StringIO(''.join(record))
        dfs.append(pd.read_table(text, sep=','))
        logger.debug("{} {}".format(rc, len(record)))
        logger.debug("{}".format(record[0]))

    return dfs


def parse_raw(in_file_name):
    """
    This function will parse a RAW file and return a PyPSA model
    """

    # Initialize output
    output = {}
    for item in modes:
        key = item['key']
        output[key] = {
            "name": item['name'],
            "key": key,
            "lines": []
        }
        # Set up column structure
        if 'columns' in item:
            output[key]['columns'] = item['columns']
            header = ",".join([c['name'] for c in item['columns']]).replace(' ', '') + '\n'
            output[key]['header'] = header
            output[key]['lines'].append(header)
        if 'records' in item:
            output[key]['records'] = item['records']
            for record in output[key]['records']:
                header = ",".join([c['name'] for c in record['columns']]).replace(' ', '') + '\n'
                record['header'] = header
        # Get parsing info
        output[key]['parse'] = item.get('parse', {})

    # Initialize header mode
    current_mode = {
        'key': 'header'
    }

    # Read file and store in container
    with open(in_file_name, 'r') as in_file:
        line = in_file.readline()
        line_num = 0
        while line:
            line_num += 1

            START = None
            STOP = None

            # Process signals
            signals = get_signals(line_num, line, current_mode)
            for signal, mode in signals:
                logger.debug("Signal: {} {} {}".format(line_num, signal['command'], mode['key']))
                if signal['command'] == 'start':
                    START = mode
                if signal['command'] == 'stop':
                    STOP = mode

            if current_mode and START and not STOP:
                raise ValueError('Current mode was never stopped')

            # Store lines
            if current_mode and (not STOP or 'keep_tail' in STOP):
                key = current_mode['key']
                output[key]['lines'].append(line)
            # print key

            if STOP and current_mode and current_mode['key'] != STOP['key']:
                raise ValueError('Attempting to stop a different mode', current_mode['key'], STOP['key'])
            elif START:
                current_mode = START
            elif STOP:
                current_mode = None

            line = in_file.readline()

    logger.debug("Captured Lines")
    for i in output:
        logger.debug('Item: {}, length: {}'.format(i, len(output[i]['lines'])))

    for i in output:
        if 'lines' not in output[i]:
            logger.debug('no lines {}'.format(i))
            continue
        lines = output[i]['lines']
        if len(lines) == 1:
            logger.debug('only header {}'.format(i))
            continue

        if 'read_table' in output[i]['parse']:
            text = StringIO(''.join(lines))
            output[i]['df'] = pd.read_table(text, sep=',')

        if 'read_transformer' in output[i]['parse']:
            output[i]['dfs'] = read_transformer(lines, output[i]['records'])
            df2 = pd.concat(output[i]['dfs'][2], axis=1, sort=False)
            df3 = pd.concat(output[i]['dfs'][3], axis=1, sort=False)
            output[i]['df'] = df3.append(df2, sort=False)

        if 'read_twodc' in output[i]['parse']:
            output[i]['dfs'] = read_twodc(lines, output[i]['records'])
            output[i]['df'] = pd.concat(output[i]['dfs'], axis=1, sort=False)

        logger.debug('{} {} {}'.format(i, len(lines), 'df' in output[i]))
        if 'df' in output[i]:
            logger.info('Parsed {} {}'.format(len(output[i]['df']), i))
            if len(lines) > 0:
                logger.debug("{}".format(lines[0]))

    return output
